from numpy.testing import assert_equal
from packaging.version import Version
import xtrack as xt
import xplt.line


def test_nominal_and_effective_order():

    def assert_order(*test_params):
        """Check elements for nominal and effective order

        Args:
            *test_params: tuples with (element, nominal_order, effective_order)
        """
        for ele, no, eo in test_params:
            assert_equal(xplt.line.nominal_order(ele), no, f"Wrong nominal order for {ele}")
            assert_equal(xplt.line.effective_order(ele), eo, f"Wrong effective order for {ele}")

    # Elements with general knl and ksl coefficients
    for cls in (xt.Multipole,):
        assert_order(
            (cls(), 0, -1),
            (cls(length=1, knl=[]), 0, -1),
            (cls(length=1, knl=[1]), 0, 0),
            (cls(length=1, knl=[0, 1]), 1, 1),
            (cls(length=1, knl=[0, 0, 1]), 2, 2),
            (cls(length=1, knl=[0, 0, 0, 1]), 3, 3),
            (cls(length=1, ksl=[]), 0, -1),
            (cls(length=1, ksl=[1]), 0, 0),
            (cls(length=1, ksl=[0, 1]), 1, 1),
            (cls(length=1, ksl=[0, 0, 1]), 2, 2),
            (cls(length=1, ksl=[0, 0, 0, 1]), 3, 3),
        )

    if Version(xt.__version__) < Version("0.60"):
        return  # elements below require newer xtrack version

    # Thick elements
    assert_order(
        (xt.Bend(length=1, k0=1), 0, 0),
        (xt.Bend(length=1, k1=1), 0, 1),
        (xt.Quadrupole(length=1, k1=1), 1, 1),
        (xt.Quadrupole(length=1, k1s=1), 1, 1),
        (xt.Sextupole(length=1, k2=1), 2, 2),
        (xt.Sextupole(length=1, k2s=1), 2, 2),
        (xt.Octupole(length=1, k3=1), 3, 3),
        (xt.Octupole(length=1, k3s=1), 3, 3),
    )

    for cls, no in (
        (xt.Bend, 0),
        (xt.Quadrupole, 1),
        (xt.Sextupole, 2),
        (xt.Octupole, 3),
    ):
        assert_order(
            (cls(), no, -1),
            (cls(length=1, knl=[]), no, -1),
            (cls(length=1, knl=[1]), no, 0),
            (cls(length=1, knl=[0, 1]), no, 1),
            (cls(length=1, knl=[0, 0, 1]), no, 2),
            (cls(length=1, knl=[0, 0, 0, 1]), no, 3),
            (cls(length=1, ksl=[]), no, -1),
            (cls(length=1, ksl=[1]), no, 0),
            (cls(length=1, ksl=[0, 1]), no, 1),
            (cls(length=1, ksl=[0, 0, 1]), no, 2),
            (cls(length=1, ksl=[0, 0, 0, 1]), no, 3),
        )

    # Special thin elements
    assert_order(
        (xt.SimpleThinBend(knl=[0]), 0, -1),
        (xt.SimpleThinBend(knl=[1]), 0, 0),
        (xt.SimpleThinQuadrupole(knl=[0, 0]), 1, -1),
        (xt.SimpleThinQuadrupole(knl=[0, 1]), 1, 1),
    )


def test_repeated_elements_survey():

    # https://github.com/xsuite/xplt/issues/31
    line = xt.Line(elements={"obm": xt.Bend(length=0.5)}, element_names=["obm", "obm"])
    # line.replace_all_repeated_elements()
    survey = line.survey()
    print(survey)

    plot = xplt.FloorPlot(survey)

    boxes = plot.artists_boxes
    assert_equal(len(boxes), 2)
    assert_equal(boxes[0].get_center(), [0.25, 0])
    assert_equal(boxes[0].get_height(), 0.5)
    assert_equal(boxes[1].get_center(), [0.75, 0])
    assert_equal(boxes[1].get_height(), 0.5)


def test_sign_sticky():

    x = [0, 1, 5, 0, 5, -3, 0, 1, 0, -2, 0]
    assert_equal(xplt.line.sign_sticky(x, initial=1), [1, 1, 1, 1, 1, -1, -1, 1, 1, -1, -1])



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import importlib
import inspect
import os, sys, shutil
from xplt import __version__


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Xplt"
copyright = "2022-2024, Philipp Niedermayer (github.com/eltos)"
author = "Philipp Niedermayer (github.com/eltos)"
github_username = "xsuite"
github_repository = "xplt"
version = __version__
release = version

# Project sources
root = os.path.abspath("..")
sys.path.insert(0, root)


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.linkcode",
    "autoapi.extension",
    "sphinx.ext.githubpages",
    "myst_nb",
    "sphinx.ext.intersphinx",
    "sphinx_codeautolink",
    "matplotlib.sphinxext.roles",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


def np_example_notebooks_init(app, *args):
    shutil.copytree(
        os.path.join(root, "examples"),
        os.path.join(app.srcdir, "examples"),
    )


def np_example_notebooks_clean(app, *args):
    shutil.rmtree(os.path.join(app.srcdir, "examples"))


def setup(app):
    app.connect("config-inited", np_example_notebooks_init)
    app.connect("build-finished", np_example_notebooks_clean)
    app.connect("autoapi-skip-member", autoapi_skip_member)


# -- Auto API ----------------------------------------------------------------
# > auto-generate API documentation

autoapi_type = "python"
autoapi_dirs = ["../xplt"]
autoapi_ignore = ["*/.ipynb_checkpoints/*"]
autoapi_options = [
    "members",
    "inherited-members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
]
autoapi_python_class_content = "both"
autoapi_member_order = "groupwise"
autodoc_typehints = "description"
autoapi_add_toctree_entry = False


def autoapi_skip_member(app, what, name, obj, skip, options):
    skip |= ".. deprecated::" in obj.docstring
    skip |= ":nodoc:" in obj.docstring
    skip |= not obj.docstring and what == "data"  # undocumented variables
    return skip


# -- MyST{NB} ----------------------------------------------------------------
# > use markdown and jupyter notebooks for building docs

myst_heading_anchors = 3
myst_enable_extensions = [
    "colon_fence",
    "dollarmath",
    "amsmath",
]
nb_render_markdown_format = "myst"
nb_execution_mode = "off"


# -- Intersphinx & Codeautolink ----------------------------------------------
# > make code examples clickable, linking to the API docs

codeautolink_concat_default = True
intersphinx_mapping = {
    #'python': ('https://docs.python.org/3', None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    #'pandas': ('https://pandas.pydata.org/pandas-docs/stable/', None),
    #'scipy': ('https://docs.scipy.org/doc/scipy/', None),
    "xsuite": ("https://xsuite.readthedocs.io/en/latest/", None),
}


# -- Linkcode ----------------------------------------------------------------
# > add [source] links to the API docs, linking to source code on GitHub


def linkcode_resolve(domain, info):
    if domain != "py" or not info["module"]:
        return None
    try:
        # inspect source
        obj = importlib.import_module(info["module"])
        for name in info["fullname"].split("."):
            obj = getattr(obj, name)
        sourcefile = inspect.getsourcefile(obj)
        sourcecode, line = inspect.getsourcelines(obj)
        # build link
        root = "xplt" + os.path.sep
        if root not in sourcefile:
            return None  # external source
        else:
            path = (root + sourcefile.split(root)[-1]).replace(os.path.sep, "/")
            filename = f"{path}#L{line}-L{line + len(sourcecode) - 1}"
            return f"https://github.com/{github_username}/{github_repository}/blob/v{version}/{filename}"
    except:
        return None


# -- HTML theme: PyData Sphinx Theme -----------------------------------------

html_theme = "pydata_sphinx_theme"
html_show_sourcelink = False
html_static_path = ["_static"]
html_css_files = ["custom.css"]
html_theme_options = {
    "show_nav_level": 2,
    "github_url": f"https://github.com/{github_username}/{github_repository}",
    "icon_links": [
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/xplt",
            "icon": "fa-solid fa-cube",
        },
    ],
    "navbar_end": ["theme-switcher", "navbar-icon-links", "version-switcher"],
    "secondary_sidebar_items": ["page-toc", "edit-this-page", "sourcelink"],
    "footer_start": ["copyright"],
    "footer_end": ["sphinx-version", "theme-version"],
    "switcher": {
        "json_url": f"https://{github_username}.github.io/{github_repository}/versions.json",
        "version_match": ".".join(version.split(".")[:2]),
    },
    "check_switcher": False,  # don't check url during build
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Methods for plotting lines"""

__author__ = "Philipp Niedermayer"
__contact__ = "eltos@outlook.de"
__date__ = "2022-11-08"

import re

import numpy as np
import matplotlib as mpl

from .util import (
    get,
    iter_elements,
    element_strength,
    defaults,
    defaults_for,
    PUBLIC_SECTION_END,
    PUBLIC_SECTION_BEGIN,
)
from .base import XPlot, XManifoldPlot
from .properties import Property, DataProperty


def nominal_order(element):
    """Get nominal element order (even if coefficients might be zero)"""
    if element.__class__.__name__ in ["Bend", "RBend"]:  # avoid using type to support Views
        return 0
    if hasattr(element, "length"):
        for n in range(10, -1, -1):
            if hasattr(element, f"k{n}") or hasattr(element, f"k{n}s"):
                return n
    if hasattr(element, "order"):
        return int(element.order)
    if hasattr(element, "knl"):
        return len(element.knl)
    return -1


def effective_order(element):
    """Get effective order of element (lowest non-zero knl or ksl strength)"""
    order = set()
    if hasattr(element, "length") and element.length > 0:
        for n in range(10):
            if get(element, f"k{n}", 0) or get(element, f"k{n}s", 0):
                order.add(n)
    for knl in ("knl", "ksl"):
        if hasattr(element, knl):
            order.update(np.flatnonzero(getattr(element, knl)).tolist())
    if order:
        return min(order)
    return -1


def order(knl):
    """Get order of knl string as int"""
    return int(re.match(r"k(\d+)l", knl).group(1))


def tanc(x):
    """Tangens cardinalis, i.e. tan(x)/x with limit tanc(0)=1"""
    # Note that np.sinc is sin(pi*x)/(pi*x) and not sin(x)/x !
    return np.sinc(x / np.pi) / np.cos(x)


def sign_sticky(arr, *, initial=1):
    """Determine sign or values, but keep previous sign for zero values"""
    s = np.sign(arr)
    if s[0] == 0:
        s[0] = initial
    index = np.arange(len(s))
    index[s == 0] = -1
    index_last_nonzero = np.maximum.accumulate(index)
    return s[index_last_nonzero]


PUBLIC_SECTION_BEGIN()


class KnlPlot(XManifoldPlot):
    """A plot for knl values along line"""

    def __init__(self, line=None, *, knl=None, filled=True, resolution="auto", **kwargs):
        """

        Args:
            line (xtrack.Line): Line of elements.
            knl (int | list[int] | str): Maximum order or list of orders n to plot knl values for.
                This can also be a manifold subplot specification string like ``"k0l+k1l,k2l"``,
                see :class:`~.base.XManifoldPlot` for details.
                If None, automatically determine from line.
            filled (bool): If True, make a filled plot instead of a line plot.
            resolution (int | string): Number of points to use for plotting. Use "auto" for
                a point at the start/end of each element.
            kwargs: See :class:`~.base.XPlot` for additional arguments

        Known issues:
            - Thin elements produced with MAD-X MAKETHIN do overlap due to the displacement introduced by the TEAPOT algorithm.
              This leads to glitches of knl being doubled or zero at element overlaps for lines containing such elements.

        """

        if knl is None:
            if line is None:
                raise ValueError("Either line or knl parameter must not be None")
            knl = int(max([effective_order(e) for e in line.elements]))
        if isinstance(knl, int):
            knl = range(knl + 1)
        if not isinstance(knl, str):
            knl = [[[f"k{n}l" for n in knl]]]
        self.resolution = resolution
        self.filled = filled

        super().__init__(on_x="s", on_y=knl, **kwargs)

        # create plot elements
        def create_artists(i, j, k, a, p):
            kwargs = dict(color=f"C{order(p)}", alpha=0.5, label=self.label_for(p, unit=True))
            if self.filled:
                kwargs.update(zorder=3, lw=0)
                artist = a.fill_between([], [], **kwargs)
                artist._constructor_kwargs = kwargs
                return artist
            else:
                return a.plot([], [], **kwargs)[0]

        self._create_artists(create_artists)

        for a in self.axflat:
            a.axhline(0, c="k", lw=1, zorder=4)
        self.legend(show="auto", ncol=5)

        # set data
        if line is not None:
            self.update(line)

    def update(self, line, *, autoscale=None):
        """
        Update the line data this plot shows

        Args:
            line (xtrack.Line): Line of elements.
            autoscale (str | None | bool): Whether and on which axes to perform autoscaling.
                One of `"x"`, `"y"`, `"xy"`, `False` or `None`. If `None`, decide based on :meth:`matplotlib.axes.Axes.get_autoscalex_on` and :meth:`matplotlib.axes.Axes.get_autoscaley_on`.


        Returns:
            changed artists
        """
        # compute knl as function of s
        if self.resolution == "auto":
            S = set()
            for name, el, s0, s1 in iter_elements(line):
                S.update({s0, s1})
            S = np.array(sorted(list(S)))
        else:
            S = np.linspace(0, line.get_length(), self.resolution)
        values = {p: np.zeros_like(S) for p in self.on_y_unique}
        for name, el, s0, s1, mask in iter_elements(line, s=S):
            for knl in self.on_y_unique:
                n = order(knl)
                values[knl][mask] += element_strength(el, n)
        if self.resolution == "auto":
            S = np.repeat(S, 2)[1:]
            values = {p: np.repeat(v, 2)[:-1] for p, v in values.items()}

        # plot
        s = S * self.factor_for("s")
        changed = []
        for i, ppp in enumerate(self.on_y):
            for j, pp in enumerate(ppp):
                ax = self.axis(i, j)
                for k, p in enumerate(pp):
                    art = self.artists[i][j][k]
                    y = self.factor_for(p) * values[p]
                    if self.filled:
                        kwargs = art._constructor_kwargs
                        changed.append(art)
                        art.remove()
                        self.artists[i][j][k] = art = ax.fill_between(S, 0, y, **kwargs)
                        art._constructor_kwargs = kwargs
                    else:
                        art.set_data((s, y))
                    changed.append(art)

                # autoscale
                self._autoscale(ax, autoscale, tight="x")

        return changed

    def prop(self, p):
        if match := re.fullmatch(r"k(\d+)l", p):
            n = match.group(1)
            return Property(symbol=f"$k_{n}l$", unit="rad" if n == "0" else f"m^-{n}")
        return super().prop(p)

    def label_for(self, *pp, unit=True, description=True):
        """
        Return axis label for list of properties, joining where possible

        Args:
            pp: Property names
            unit (bool): Whether to include unit
            description (bool): Whether to include description

        Returns:
            str: Axis label
        """
        if len(pp) > 1 and np.all([re.match(r"k\d+l", p) for p in pp]):
            label = "$k_nl$"
            if unit:
                label += " / $m^{-n}$"
            return label
        return super().label_for(*pp, unit=unit, description=description)


class FloorPlot(XPlot):
    """A floor plan of the line based on survey data"""

    def __init__(
        self,
        survey=None,
        line=None,
        projection="ZX",
        *,
        default_boxes=None,
        boxes=None,
        labels=False,
        ignore=None,
        element_width=1,
        axis="equal",
        **kwargs,
    ):
        """

        Args:
            survey (Any | None): Survey data from MAD-X or Xsuite.
            line (None | xtrack.Line): Optional Xsuite line object for backwards compatible coloring of Multipoles (deprecated).
            projection (str): The projection to use: A pair of coordinates ('XZ', 'ZY' etc.)
            boxes (None | bool | str | iterable | dict): Config option for showing colored boxes for elements. See below.
                Detailed options can be "length" and all options suitable for a patch, such as "color", "alpha", etc.
                By default, all elements with a multipole order are shown, or if line is None then all element.
            default_boxes (bool): Whether to keep the default boxes even if custom box options are specified.
            labels (None | bool | str | iterable | dict): Config option for showing labels for elements. See below.
                Detailed options can be "text" (e.g. "Dipole {name}" where name will be
                replaced with the element name) and all options suitable for an annotation,
                such as "color", "alpha", etc.
            ignore (None | str | list[str]): Optional patter or list of patterns to ignore elements with matching names.
                Note that drift spaces are always ignored.
            element_width (float): Width of element boxes.
            axis (str): Aspect ratio of the plot. Default is 'equal'.
            kwargs: See :class:`~.base.XPlot` for additional arguments


        The config options passed to boxes and labels can be:
            - None: Use good defaults.
            - A bool: En-/disable option for all elements (except drifts).
            - A str (regex): Filter by element name.
            - A list, tuple or numpy array: Filter by any of the given element names
            - A dict: Detailed options to apply for each element in the form of
              `{"regex": {...}}`. For each matching element name, the options are used.

        """

        kwargs["_properties"] = defaults(
            kwargs.get("_properties"),
            X=DataProperty("X", "m"),
            Y=DataProperty("Y", "m"),
            Z=DataProperty("Z", "m"),
            theta=DataProperty("theta", "rad", "$\\Theta$"),
            phi=DataProperty("phi", "rad", "$\\Phi$"),
            psi=DataProperty("psi", "rad", "$\\Psi$"),
        )

        super().__init__(**kwargs)

        if projection == "3D":
            raise NotImplementedError()

        self.projection = projection
        self.boxes = boxes
        self.default_boxes = default_boxes if default_boxes is not None else (boxes is not False)
        self.labels = labels
        self.ignore = [ignore] if isinstance(ignore, str) else ignore
        self.element_width = element_width

        if isinstance(self.boxes, (list, tuple, np.ndarray)):
            self.boxes = "|".join(["^" + ll + "$" for ll in self.boxes])
        if isinstance(self.labels, (list, tuple, np.ndarray)):
            self.labels = "|".join(["^" + ll + "$" for ll in self.labels])

        # Create plot
        self.ax.set(
            xlabel=self.label_for(self.projection[0]), ylabel=self.label_for(self.projection[1])
        )
        self.ax.axis(axis)

        # create plot elements
        (self.artist_beamline,) = self.ax.plot([], [], "k-")
        self.artist_startpoint = mpl.patches.FancyArrowPatch(
            (0, 0), (0, 0), mutation_scale=20, color="k", arrowstyle="-|>", zorder=5, lw=0
        )
        self.ax.add_patch(self.artist_startpoint)
        self.artists_boxes = []
        self.artists_labels = []

        if survey is not None:
            self.update(survey, line)

    def update(self, survey=None, line=None, *, autoscale=None):
        """
        Update the survey data this plot shows

        Args:
            survey (Any | None): Survey data. Defaults to `line.survey()`.
                For convenience, passing line as first argument is also supported.
            line (None | xtrack.Line): Optional Xsuite line object for backwards compatible coloring of Multipoles (deprecated).
            autoscale (str | None | bool): Whether and on which axes to perform autoscaling.
                One of `"x"`, `"y"`, `"xy"`, `False` or `None`. If `None`, decide based on :meth:`matplotlib.axes.Axes.get_autoscalex_on` and :meth:`matplotlib.axes.Axes.get_autoscaley_on`.

        Returns:
            changed artists

        """

        changed = []

        if self.projection == "3D":
            ...
            raise NotImplementedError()

        else:
            A, B = self.projection
            scale = self.factor_for(A)
            if scale != self.factor_for(B):
                # can't handle this, because angles are not preserved
                raise ValueError(f"Display units for {A} and {B} must be equal!")

            X = get(survey, A) * scale  # coordinate to be plotted on x-axis
            Y = get(survey, B) * scale  # coordinate to be plotted on y-axis

            # rotation of element (angle between plot x-axis and tangential beam direction)
            if self.projection in ("ZX", "XZ"):
                RT = get(survey, "theta")
            elif self.projection in ("ZY",):
                RT = get(survey, "phi")
            else:
                raise ValueError(f"Unknown projection {self.projection}")

            # ang: function to transform angles from data (A-B) to axis (X-Y) coordinate system
            if self.projection in ("ZX", "ZY"):

                def ang(a):
                    return a

            else:

                def ang(a):
                    return np.pi / 2 - a

            NAME = get(survey, "name")
            ARC = -np.round(np.diff(RT, append=RT[-1]), 9)  # bending angle to next element
            HELICITY = sign_sticky(ARC)  # if the beamline is left or right bending
            RR = ang(
                RT - ARC / 2 + HELICITY * np.pi / 2
            )  # angle between plot x-axis and radial vector of bending

            LENGTH = get(survey, "length", np.zeros_like(NAME))
            IS_THICK = get(survey, "isthick", np.zeros_like(NAME))
            ORDER = get(survey, "order", -np.ones_like(NAME))
            if (TYPE := get(survey, "element_type", None)) is not None:
                # map element type to order when order is not in survey
                for type, o in {
                    "Bend": 0,
                    "RBend": 0,
                    "Quadrupole": 1,
                    "Sextupole": 2,
                    "Octupole": 3,
                    "Multipole": 999,
                }.items():
                    ORDER[(ORDER < 0) & (TYPE == type)] = o

            # beam line
            ############
            self.artist_beamline.set_data(X, Y)
            changed.append(self.artist_beamline)
            # start point arrow
            i = np.argmax((X > X[0]) | (Y > Y[0]))
            self.artist_startpoint.set_positions((2 * X[0] - X[i], 2 * Y[0] - Y[i]), (X[0], Y[0]))
            changed.append(self.artist_startpoint)

            # elements
            ###########
            while len(self.artists_boxes) > 0:
                # remove old artists
                self.artists_boxes.pop().remove()

            while len(self.artists_labels) > 0:
                # remove old artists
                self.artists_labels.pop().remove()

            legend_entries = []
            for i, (x, y, rt, name, arc) in enumerate(zip(X, Y, RT, NAME, ARC)):
                helicity, rr, length, is_thick, order = (
                    HELICITY[i],
                    RR[i],
                    LENGTH[i],
                    IS_THICK[i],
                    ORDER[i],
                )

                if name == "_end_point":
                    continue

                element = None
                if line is not None:
                    # Fallback to extract missing properties from line
                    try:
                        element = line.get(
                            name
                        )  # required also for custom text formatting if user wants to
                        if not is_thick:
                            is_thick = element.isthick
                        if order < 0 or order > 100:
                            order = effective_order(element)
                        if not length:
                            length = get(element, "length", 0)
                    except (TypeError, KeyError):
                        pass

                if self.ignore is not None:
                    if np.any([re.match(pattern, name) is not None for pattern in self.ignore]):
                        continue  # skip ignored

                # box
                ######

                # default style
                default_box_style = dict(
                    color=f"C{order}" if order >= 0 else "k",
                    length=length or 0,
                    label={
                        0: "Bending magnet" if arc else None,
                        1: "Quadrupole magnet",
                        2: "Sextupole magnet",
                        3: "Octupole magnet",
                        999: "Multipole magnet",
                    }.get(order),
                )

                boxes = self.boxes
                box_style = self._get_config(boxes, name, **default_box_style)
                if box_style is None and self.default_boxes and order >= 0:
                    box_style = default_box_style

                if box_style is not None:
                    width = box_style.pop("width", self.element_width) * scale
                    length = box_style.pop("length", 0) * scale
                    if box_style.get("label") in legend_entries:
                        box_style.pop("label")  # prevent duplicate legend entries
                    else:
                        legend_entries.append(box_style.get("label"))

                    # Handle thick elements
                    if is_thick:
                        # Find the center of single kick for equivalent thin element
                        d = length * tanc(arc / 2) / 2
                        x += d * np.cos(ang(rt))
                        y += d * np.sin(ang(rt))

                    if length > 0 and arc:
                        # bending elements as wedge
                        rho = length / arc
                        x_center = x - helicity * rho * np.cos(rr) / np.cos(arc / 2)
                        y_center = y - helicity * rho * np.sin(rr) / np.cos(arc / 2)
                        wedge_kwargs = defaults_for(
                            mpl.patches.Wedge,
                            box_style,
                            center=(x_center, y_center),
                            r=rho + width / 2,
                            width=width,
                            theta1=np.rad2deg(rr - helicity * arc / 2)
                            + 90 * (1 - helicity),  # rr - arc/2),
                            theta2=np.rad2deg(rr + helicity * arc / 2)
                            + 90 * (1 - helicity),  # rr + arc/2),
                            alpha=0.5,
                            zorder=3,
                        )
                        box = mpl.patches.Wedge(**wedge_kwargs)
                    else:
                        # other elements as rect
                        box = mpl.patches.Rectangle(
                            **defaults_for(
                                mpl.patches.Rectangle,
                                box_style,
                                xy=(x - width / 2, y - length / 2),
                                width=width,
                                height=length,
                                angle=np.rad2deg(ang(rt - arc / 2)) - 90,
                                rotation_point="center",
                                alpha=0.5,
                                zorder=3,
                            )
                        )
                    self.ax.add_patch(box)
                    self.artists_boxes.append(box)
                    changed.append(box)

                # label
                ########

                labels = self.labels
                if labels is None:
                    labels = order >= 0
                label_style = self._get_config(labels, name, text=name)

                if label_style is not None:
                    width = label_style.pop("width", self.element_width * scale)
                    label_style["text"] = label_style["text"].format(
                        name=name, length=length, arc=arc, order=order, element=element
                    )

                    label = self.ax.annotate(
                        **defaults_for(
                            "text",
                            label_style,
                            xy=(x, y),
                            xytext=(x + 1.5 * width * np.cos(rr), y + 1.5 * width * np.sin(rr)),
                            # xytext=(40*np.cos(rr), 40*np.sin(rr)),
                            # textcoords='offset points',
                            va={1: "bottom", 0: "center", -1: "top"}[np.round(np.sin(rr))],
                            ha={1: "left", 0: "center", -1: "right"}[np.round(np.cos(rr))],
                            # rotation=(np.rad2deg(rr)+90)%180-90,
                            arrowprops=dict(arrowstyle="-", color="0.5", shrinkB=5),
                            clip_on=True,
                            zorder=5,
                        )
                    )
                    self.artists_labels.append(label)
                    changed.append(label)

            # autoscale
            self._autoscale(
                self.ax,
                autoscale,
                artists=self.artists_boxes + self.artists_labels + [self.artist_beamline],
            )

        return changed

    def legend(self, **kwargs):
        self.ax.legend(**kwargs)

    def add_scale(self, scale=None, label=None, *, loc="auto", color="k", fontsize="x-small"):
        """Add a scale patch (a yardstick or ruler)

        Args:
            scale (float): The length of the scale in data units (typically meter).
            label (str | None): A label for the scale.
            loc (str): The location of the scale. Can be any of the usual matplotlib locations,
                e.g. 'auto', 'upper left', 'upper center', 'upper right', 'center left', 'center',
                'center right', 'lower left', 'lower center, 'lower right'.
            color (Any): Color for the patch.
            fontsize (Any): Font size of the label.

        Returns:
            matplotlib.offsetbox.AnchoredOffsetbox: The artist added
        """
        if scale is None:
            scale = 5
        else:
            # convert to display units
            scale = scale * self.factor_for(self.projection[0])

        if label is None:
            unit = self.display_unit_for(self.projection[0])
            label = f"{scale:g} {unit}"

        return super().add_scale(
            self.ax, scale, label=label, loc=loc, color=color, fontsize=fontsize
        )

    @staticmethod
    def _get_config(config, name, **default):
        if isinstance(config, str):
            if re.match(config, name):
                return default
        elif isinstance(config, dict):
            for pattern, args in config.items():
                if re.match(pattern, name):
                    if args is False:
                        return None
                    if isinstance(args, dict):
                        return defaults(args, **default)
                    return default
        elif config:
            return default


__all__ = PUBLIC_SECTION_END()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Base methods for plotting"""

__author__ = "Philipp Niedermayer"
__contact__ = "eltos@outlook.de"
__date__ = "2022-11-08"


import re
import types
import uuid

import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.scale
import matplotlib.transforms
import matplotlib.patches
import matplotlib.path
import numpy as np

from .util import defaults, flattened, defaults_for, AUTO
from .properties import (
    Property,
    find_property,
    DataProperty,
    arb_unit,
    _convert_value_to_unit,
    _fmt_qty,
    _has_pint,
)


class ManifoldMultipleLocator(mpl.ticker.MaxNLocator):
    def __init__(self, fixed_multiples, n=5, minor_n=None):
        """A multiple locator that chooses its base from a set of multiples to yield about n ticks

        For ranges smaller than the smallest fixed_multiple, the default MaxNLocator is used
        For ranges larger than the largest fixed_multiple, a multiple of the later is used

        Args:
            fixed_multiples (list of float): multiples to choose from
            n (int): number of ticks to produce (best effort)
            minor_n (list of float): If given, produce that many minor ticks in between each major tick. Length must match the length of fixed_multiples
        """
        super().__init__(n)
        self.fixed_multiples = fixed_multiples
        self.n = n
        self.minor_n = minor_n

    def _raw_ticks(self, vmin, vmax):
        if vmax - vmin < self.n * self.fixed_multiples[0]:
            return super()._raw_ticks(vmin, vmax)
        for step in self.fixed_multiples:
            if (vmax - vmin) / step <= self.n:
                break
        while (vmax - vmin) / step > self.n:
            step += self.fixed_multiples[-1]
        if self.minor_n is not None:
            if step in self.fixed_multiples:
                step /= self.minor_n[self.fixed_multiples.index(step)]
            else:
                step /= self.minor_n[-1]
        return np.arange(int(vmin / step) * step, vmax + step, step)


class AngleLocator(ManifoldMultipleLocator):
    def __init__(self, minor=False, deg=True):
        """A tick locator for angles

        Args:
            minor (bool): If true, return a minor locator. By default a major locator is returned.
            deg (bool): If true, locate angles is degree. If false, in radians.
        """
        multiples = (5, 15, 30, 45, 60, 90, 120, 180, 360)
        if not deg:
            multiples = list(np.deg2rad(multiples))
        subdivide = (5, 3, 3, 3, 4, 3, 4, 4, 4)
        super().__init__(multiples, 5, subdivide if minor else None)


class RadiansFormatter(mpl.ticker.Formatter):
    """A tick formatter to format angles in radians as fractions or multiples of pi"""

    def __call__(self, x, pos=None):
        if x == 0:
            return "0"
        s = "-" if x < 0 else ""
        x = abs(x)
        if x == np.pi:
            return f"${s}\\pi$"
        for n in (2, 3, 4, 6, 8, 12):
            m = round(x / (np.pi / n))
            if abs(x - m * np.pi / n) < 1e-10 and m / n != m // n:
                if m == 1:
                    m = ""
                return f"${s}{m}\\pi/{n}$"
        return f"${x/np.pi:g}\\pi$"


class TwinFunctionLocator(mpl.ticker.Locator):
    def __init__(self, twin_locator, function_twin_to_this, function_this_to_twin, granularity=1):
        """A locator for twin axes with non-linear functional dependence

        Finds nice tick locations close to the twin, but uses a custom function to place ticks at integer multiples of granularity.
        This is useful for twin axes which share the same limits, but are formatted with values based on different functions.

        Args:
            twin_locator (mpl.ticker.Locator): The other locator to align tick locations with
            function_twin_to_this (function): Function to calculate tick values of this axis given the tick values of the other axis.
            function_this_to_twin (function): Function to calculate tick values of the other axis given the tick values of this axis.
            granularity (float): Base at multiples of which to locate ticks.
        """
        self.twin_locator = twin_locator
        self.twin2this = function_twin_to_this
        self.this2twin = function_this_to_twin
        self.granularity = granularity

    def __call__(self):
        """Return the locations of the ticks."""
        return self.tick_values(*self.axis.get_view_interval())

    def tick_values(self, vmin, vmax):
        twin_values = np.array(self.twin_locator.tick_values(vmin, vmax))
        this_values = self.twin2this(twin_values)
        this_values = np.round(this_values / self.granularity) * self.granularity
        twin_values = self.this2twin(this_values)
        twin_values = twin_values
        return twin_values


class TransformedLocator(mpl.ticker.Locator):
    def __init__(
        self, locator, transform=lambda x: x, inverse=lambda x: x, vmin=-np.inf, vmax=np.inf
    ):
        """A transformed locator with non-linear functional wrappers and limits

        Clips ticks to limits and then transforms the values before calling the dependent locator.

        Args:
            locator (mpl.ticker.Locator): The dependent locator to use for actual tick locating
            transform (function): Function to transform tick values of this locator to the values of the dependent locator
            inverse (function): Inverse of transform
            vmin (float): Optional lower limit for ticks in this locator
            vmax (float): Optional upper limit for ticks in this locator
        """
        self.locator = locator
        self.transform = transform
        self.inverse = inverse
        self.vmin = vmin
        self.vmax = vmax

    def __call__(self):
        """Return the locations of the ticks."""
        return self.tick_values(*self.axis.get_view_interval())

    def tick_values(self, vmin, vmax):
        # clip and transform limits
        vmin = self.transform(max(vmin, self.vmin))
        vmax = self.transform(min(vmax, self.vmax))
        # locate ticks
        ticks = self.locator.tick_values(vmin, vmax)
        # inverse transform
        ticks = self.inverse(ticks)
        return ticks


class DiscontinuousLinearScale(mpl.scale.LinearScale):
    name = "discontinuous_linear"

    def __init__(self, axis, *, breaks=(), space=0, hide=False):
        super().__init__(axis)
        self.breaks = breaks
        self.space = space
        self.hide = hide

    def set_default_locators_and_formatters(self, axis):
        axis.set(
            major_locator=DiscontinuousLocator(self.breaks, self.space),
            major_formatter=DiscontinuousFormatter(self.breaks),
        )
        self.add_discontinuity_markers(axis)

    def get_transform(self):
        return DiscontinuousTransform(self.breaks, self.space, hide=self.hide)

    def add_discontinuity_markers(self, axis):
        ax = axis.axes
        xy = lambda x, y: (x, y) if axis == ax.xaxis else (y, x)
        trans = mpl.transforms.blended_transform_factory(*xy(ax.transData, ax.transAxes))
        kwargs = dict(color="k", linewidth=0.8, transform=trans, clip_on=False, zorder=9)
        for br in self.breaks:
            if self.hide:
                ax.add_artist(
                    mpl.patches.Rectangle(xy(br[0], 0), *xy(br[1] - br[0], 1), fc="w", **kwargs)
                )
            else:
                for i in range(2):
                    x, w, h = np.mean(br), np.diff(br).item() / 4, 0.02
                    points = list(zip(*xy((x - w, x, x + w, x), (i - h, i + h, i + h, i - h))))
                    ax.add_artist(
                        mpl.patches.PathPatch(mpl.path.Path(points), ec="none", fc="w", **kwargs)
                    )
                    ax.add_artist(
                        mpl.patches.PathPatch(
                            mpl.path.Path(
                                points,
                                [
                                    mpl.path.Path.MOVETO,
                                    mpl.path.Path.LINETO,
                                    mpl.path.Path.MOVETO,
                                    mpl.path.Path.LINETO,
                                ],
                            ),
                            **kwargs,
                        )
                    )


class DiscontinuousLocator(mpl.ticker.AutoLocator):
    def __init__(self, breaks, space, **kwargs):
        super().__init__(**kwargs)
        self.breaks = breaks
        self.space = space

    def set_params(self, **kwargs):
        if "nbins" in kwargs:
            self._nbins_real = kwargs.pop("nbins")
        super().set_params(**kwargs)

    def _raw_ticks(self, vmin, vmax):
        visible = total = vmax - vmin
        for br in self.breaks:
            left, right = np.clip(br, vmin, vmax)
            visible -= right - left - self.space

        nbins = self._nbins_real
        if nbins == "auto":
            if self.axis is not None:
                nbins = np.clip(self.axis.get_tick_space(), max(1, self._min_n_ticks - 1), 9)
            else:
                nbins = 9

        self._nbins = nbins * total / visible
        ticks = super()._raw_ticks(vmin, vmax)

        # Remove ticks inside breaks
        # for left, right in self.breaks:
        #    ticks = ticks[(ticks <= left) | (ticks >= right)]

        return ticks


class DiscontinuousFormatter(mpl.ticker.ScalarFormatter):
    def __init__(self, breaks, **kwargs):
        super().__init__(**kwargs)
        self.breaks = breaks

    def __call__(self, x, pos=None):
        """Return the format for tick value *x* at position *pos*."""
        for left, right in self.breaks:
            if left < x < right:
                return None
        return super().__call__(x, pos)


class DiscontinuousTransform(mpl.transforms.Transform):
    input_dims = output_dims = 1

    def __init__(self, breaks, space, *, hide=True):
        super().__init__()
        self.breaks = breaks
        self.spaces = space * (np.ones(len(breaks)) if np.isscalar(space) else 1)
        self.hide = hide

    def transform_non_affine(self, values):
        a = np.asanyarray(values, dtype="float").copy()
        breaks = np.asanyarray(self.breaks, dtype="float").copy()
        spaces = np.asanyarray(self.spaces, dtype="float")
        m = np.zeros_like(a)
        for (l, r), s in zip(breaks, spaces):
            gap, over = (a > l) & (a < r), a >= r
            m[gap] = 1
            a[gap] = l + s * (a[gap] - l) / (r - l)
            a[over] -= r - l - s
            breaks[breaks >= r] -= r - l - s
        if self.hide:
            a = np.ma.masked_where(m, a)
        return a

    def inverted(self):
        spaces = np.diff(self.breaks, axis=1)[:, 0]
        breaks = self.transform_non_affine(self.breaks)
        return DiscontinuousTransform(breaks, spaces, hide=False)


class XPlot:
    def __init__(
        self,
        *,
        data_units=None,
        display_units=None,
        ax=None,
        grid=True,
        log=False,
        nntwins=None,
        annotation=None,
        _properties=None,
        **subplots_kwargs,
    ):
        """
        Base class for plotting

        Initialize the subplots, axes and twin axes

        Args:
            data_units (dict | None): Units of the data. If None, the units are determined from default and user property settings.
            display_units (dict | None): Units to display the data in. If None, the units are determined from the data.
            ax (matplotlib.axes.Axes | None): Axes to plot onto. If None, a new figure is created.
            grid (bool): If True, show grid lines on all axes.
            log (bool | str): If True, `"xy"`, `"x"` or `"y"`, make the respective axis/axes log-scaled
            nntwins (list | None): List defining how many twin axes to create for each subplot.
            annotation (bool | None): Whether to add an annotation or not. If None (default) add it unless `ax` is passed.
            subplots_kwargs: Keyword arguments passed to :func:`matplotlib.pyplot.subplots` command when a new figure is created.
        """

        self._default_properties = {}
        if _properties:
            self._default_properties.update(_properties)
        self._user_properties = {}
        if data_units:
            for name, arg in data_units.items():
                self._user_properties[name] = (
                    DataProperty(name, arg) if isinstance(arg, str) else arg
                )
        self._display_units = defaults(
            display_units, **dict(s="m", x="mm", y="mm", p="mrad") if _has_pint() else {}
        )

        if annotation is None:
            annotation = ax is None

        # Create plot axes
        if ax is None:
            fig, ax = plt.subplots(**subplots_kwargs)
        self.ax = ax
        self.fig = self.axflat[0].figure
        self.axflat_twin = []

        for i, a in enumerate(self.axflat):
            if grid:
                a.grid(grid, alpha=0.5)
            for xy in "xy":
                if log is True or (isinstance(log, str) and xy in log.lower()):
                    a.set(**{f"{xy}scale": "log"})

            # Create twin axes
            self.axflat_twin.append([])
            if nntwins is not None:
                for j in range(nntwins[i]):
                    twin = a.twinx()
                    if log is True or (isinstance(log, str) and "y" in log.lower()):
                        twin.set(yscale="log")
                    try:
                        # hack for shared cyclers
                        # see https://github.com/matplotlib/matplotlib/issues/19479
                        twin._get_lines = a._get_lines
                    except:
                        print(
                            "Warning: failed to share cyclers, please manually set trace colors for twin axes and make sure to upvote https://github.com/matplotlib/matplotlib/issues/19479"
                        )
                        pass  # ignore
                    if j > 0:
                        twin.spines.right.set_position(("axes", 1 + 0.2 * j))
                    self.axflat_twin[i].append(twin)

        # Create annotation
        if annotation:
            self.annotation = self.fig.text(
                0.005, 0.005, "", ha="left", va="bottom", c="gray", linespacing=1, fontsize=8
            )
        else:
            self.annotation = None

    def _autoscale(
        self, ax, axis="xy", *, artists=None, data=None, reset=False, freeze=True, tight=None
    ):
        """Autoscale axes to fit given artists

        If neither artists nor data is specified, consider all artists associated with the axis.

        Args:
            ax (matplotlib.axes.Axes): Axes to autoscale
            axis (str | None | bool): Whether and on which axes to perform autoscaling.
                One of `"x"`, `"y"`, `"xy"`, `False` or `None`. If `None`, decide based on :meth:`matplotlib.axes.Axes.get_autoscalex_on` and :meth:`matplotlib.axes.Axes.get_autoscaley_on`.
                For backwards compatibility, the following aliases are also supported: `"both"`, `True`, `""`.
            artists (iterable): Artists to consider (if any)
            data (iterable): Data points to consider (if any) in the form [(x1,y1), (x2,y2), ...]
            reset (bool): Whether to ignore any data limits already registered.
            freeze (bool): Whether to keep the updated axes limits (True) or enable automatic
                autoscaling on future draws (for all present and new artists).
            tight (str | None | bool): Enables tight scaling without margins for axis.
                Any of `"x"`, `"y"`, `"xy"` or `"both"`, `""` or `None`.

        Returns:
            str: The axes autoscaled (`""`, `"x"`, `"y"` or `"xy"`)
        """
        if axis is None:
            axis = "x" * ax.get_autoscalex_on() + "y" * ax.get_autoscaley_on()
        axis = {False: "", True: "xy", "both": "xy"}.get(axis, axis)
        if not axis:
            return ""
        tight_x = ("x" in tight or tight == "both") if isinstance(tight, str) else tight
        tight_y = ("y" in tight or tight == "both") if isinstance(tight, str) else tight

        if artists is None and data is None:
            # use limits from all artists associated with the axis   https://stackoverflow.com/a/71966295
            # unlike ax.relim(), we also handle collections
            artists = ax.lines + ax.collections + ax.patches + ax.images

        data = [] if data is None else data[:]  # make a copy so we can safely append
        limits = []

        # Get data limits from artists
        drawn = False
        for art in flattened(artists) if artists is not None else []:
            if hasattr(art, "get_datalim"):
                # Collections, patches, etc.
                lim = art.get_datalim(ax.transData)
                if not np.all(np.isfinite(lim)):
                    # fallback to offsets (e.g. for hexbin)
                    data.extend(art.get_offsets())
                else:
                    limits.append(lim)

            elif hasattr(art, "get_data"):
                # e.g. Line2D
                data.extend(np.transpose(art.get_data()))

            elif isinstance(art, mpl.artist.Artist):
                # Any other artist (e.g. Text), but requires to draw figure
                if not drawn:
                    self.fig.canvas.draw()
                    drawn = True
                lim = art.get_window_extent().transformed(ax.transData.inverted())
                limits.append(lim)

            elif art is not None:
                raise NotImplementedError(f"Autoscaling not implemented for {art!r}")

        # Add limits from raw data
        if len(data) > 0:
            x, y = [a[np.isfinite(a)] for a in np.transpose(data)]
            limits.append(mpl.transforms.Bbox.from_extents(x.min(), y.min(), x.max(), y.max()))

        # Update axes limits
        if len(limits) > 0:
            dataLim = mpl.transforms.Bbox.union(limits)

            if reset:
                ax.dataLim = dataLim
            else:
                ax.update_datalim(dataLim)  # takes previous datalim into account

        # Autoscale (on next and future draws)
        if "x" in axis:
            if tight_x is False and ax.margins()[0] == 0:
                ax.margins(x=0.05)  # restore default margins
            ax.autoscale(axis="x", tight=tight_x)
        if "y" in axis:
            if tight_y is False and ax.margins()[1] == 0:
                ax.margins(y=0.05)  # restore default margins
            ax.autoscale(axis="y", tight=tight_y)

        if freeze:
            # perform autoscale immediately and freeze limits
            ax.autoscale_view()
            ax.set(xlim=ax.get_xlim(), ylim=ax.get_ylim())
            ax.set_autoscale_on(False)

        return axis

    def annotate(self, text, **kwargs):
        if self.annotation is not None:
            self.annotation.set(text=text, **kwargs)

    @property
    def axflat(self):
        """Return a flat list of all primary axes"""
        return flattened(self.ax)

    def axis(self, subplot=0, twin=0):
        """Return the axis for a given flat subplot index and twin index

        Args:
            subplot (int): Flat subplot index
            twin (int): Twin index

        Returns:
            matplotlib.axes.Axes: Axis for the given subplot and twin index
        """
        return self.axflat_twin[subplot][twin - 1] if twin else self.axflat[subplot]

    def axes(self, subplots="all", twins="all"):
        """Generator yielding axes for the given flat subplot and twin indices

        Args:
            subplots (int | list[int] | str): Flat subplot indices or ``"all"``
            twins (int | list[int] | str | bool): Twin index or indices or ``"all"``.
                ``True`` as alias for ``"all"`` and ``False`` as alias for ``0`` are also supported.

        Yields:
            matplotlib.axes.Axes: Iterator over the selected axes where
        """
        subplots = [subplots] if isinstance(subplots, int) else subplots
        twins = [twins] if isinstance(twins, int) else {False: [0], True: "all"}.get(twins, twins)
        for s in range(len(self.axflat)):
            if subplots != "all" and s not in subplots:
                continue
            for t in range(len(self.axflat_twin[s]) + 1):
                if twins != "all" and t not in twins:
                    continue
                yield self.axflat_twin[s][t - 1] if t else self.axflat[s]

    def save(self, fname, **kwargs):
        """Save the figure with sensible default options

        Args:
            fname (str): Filename
            kwargs: Keyword arguments passed to :meth:`matplotlib.figure.Figure.savefig`
        """
        self.fig.savefig(fname, **defaults(kwargs, dpi=300))

    def title(self, title, **kwargs):
        """Set figure title

        Args:
            title (str): Title
            kwargs: Keyword arguments passed to :meth:`matplotlib.figure.Figure.suptitle`
        """
        self.fig.suptitle(title, **kwargs)

    def factor_for(self, p):
        """Return factor to convert parameter into display unit

        Args:
            p (str): Property name

        Returns:
            float: Factor to convert parameter into display unit
        """
        return _convert_value_to_unit(1, self.prop(p).unit, self.display_unit_for(p))

    def display_unit_for(self, p):
        """Return display unit for parameter

        Args:
            p (str): Property name

        Returns:
            str: Display unit
        """
        if p is None:
            return None

        if p in self._display_units:
            return self._display_units[p]

        if len(p) > 1 and p[-1] in "xy":
            if p[:-1] in self._display_units:
                return self._display_units[p[:-1]]

        return self.prop(p).unit  # default to data unit

    def prop(self, name):
        """Get property by key

        Args:
            name (str): Key
        Returns:
            Property: The property
        """
        prop = find_property(
            name,
            extra_user_properties=self._user_properties,
            extra_default_properties=self._default_properties,
        )
        return prop.with_property_resolver(self.prop)

    def _legend_label_for(self, p):
        """
        Return legend label for a single property

        Args:
            p (str): Property name

        Returns:
            str: Legend label
        """
        return self.prop(p).description or self._symbol_for(p)

    def _axis_label_for(self, p, description=True):
        """Return axis label for a single property

        The label is without unit, as the unit is added in `label_for` separately.

        Args:
            p (str): Property name
            description (bool): Whether to include the description (if any)

        Returns:
            str: Axis label
        """
        label = self._symbol_for(p)
        if description and (desc := self.prop(p).description):
            label = desc + "   " + label
        return label

    def _symbol_for(self, p):
        """Return symbol for a single property

        This method can be overridden by subclasses to provide custom symbols,
        such as an average over a quantity etc.

        Args:
            p (str): Property name

        Returns:
            str: Symbol
        """
        prop = self.prop(p)
        return prop.symbol

    def label_for(self, *pp, unit=True, description=True):
        """
        Return axis label for list of properties, joining where possible

        Args:
            pp: Property names
            unit (bool): Whether to include unit
            description (bool): Whether to include description

        Returns:
            str: Axis label
        """

        # filter out None
        pp = [p for p in pp if p is not None]

        if len(pp) == 0:
            return ""

        # if there are different units, treat them separately
        units = np.array([self.display_unit_for(p) for p in pp])
        if unit and np.unique(units).size > 1:
            # different units, treat parameters for each unit separately
            lines = []
            for u in np.unique(units):
                pp_with_unit = np.array(pp)[units == u]
                lines.append(self.label_for(*pp_with_unit, unit=unit))
            return "\n".join(lines)

        # reduce complexity, by combining labels with common prefix
        labels, combined = [], {}

        for p in pp:
            label = self._axis_label_for(p, description)

            if m := re.fullmatch("\\$(.+)_(.)\\$", label):
                pre, suf = m.groups()
                if pre not in combined:
                    combined[pre] = []
                combined[pre].append(suf)
            else:
                labels.append(label)

        for pre, ss in combined.items():
            s = ",".join(ss)
            labels.append(f"${pre}_{{{s}}}$")

        # build label
        label = ", ".join(np.unique(labels))

        # add unit
        if unit:
            append = None
            if units[0] == arb_unit:  # arbitrary unit
                append = " / " + arb_unit
            else:
                display_unit = units[0]  # all have the same unit (see above)
                if display_unit and display_unit != "1":
                    append = f" / ${_fmt_qty(None, display_unit)}$"  # see "NIST Guide to the SI"
            if append:
                # heuristic: if labels contain expressions with +, - or ± then add parentheses
                if re.findall(r"([-+±]|\\pm)(?![^(]*\))(?![^{]*\})", label.split("   ")[-1]):
                    label = f"({label})"
                label += append

        return label

    @staticmethod
    def _set_axis_ticks_angle(yaxis, minor=True, deg=False):
        """Set ticks locator and formatter to display an angle

        This will set ticks at multiples or fractions of 180° or pi with appropriate labels

        Args:
            yaxis (matplotlib.axis.Axis): The axis to format (ax.xaxis or ax.yaxis)
            minor (bool): If true (default), also set the minor locator
            deg (bool): If true, use angles in degree. If false (default), in radians.
        """
        yaxis.set_major_locator(AngleLocator(deg=deg))
        if minor:
            yaxis.set_minor_locator(AngleLocator(deg=deg, minor=True))
        if not deg:
            yaxis.set_major_formatter(RadiansFormatter())

    @staticmethod
    def plot_harmonics(
        ax, v, dv=0, *, n=20, scale_width=True, vertical=True, inverse=False, **plot_kwargs
    ):
        """Add vertical lines or spans indicating the location of values or spans and their harmonics

        Indicates the bands at the h-th harmonics (h = 1, 2, ..., n) at
            - h * (v ± dv/2)      if scale_width and not inverse (default)
            - h * v ± dv/2        if not scale_width and not inverse
            - h / (v ± dv/2)      if inverse and scale_width
            - 1 / ( v/h ± dv/2 )  if inverse and not scale_width

        Args:
            ax (matplotlib.axes.Axes): Axes to plot onto.
            v (float | list[float]): Value or list of values.
            dv (float | list[float]): Width or list of widths centered around value(s).
            n (int): Number of harmonics to plot.
            scale_width (bool): Whether to scale the width for higher harmonics or keep it constant.
            vertical (bool): Plot vertical lines if true, horizontal otherweise.
            inverse (bool): If true, plot harmonics of n/(v±dv) instead of n*(v±dv).
                Useful to plot frequency harmonics in time domain and vice-versa.
            plot_kwargs: Keyword arguments to be passed to plotting method
        """
        if not hasattr(v, "__iter__"):
            v = [v]
        if not hasattr(dv, "__iter__"):
            dv = [dv] * len(v)
        kwargs = defaults(plot_kwargs, zorder=1.9, color="gray", lw=1)
        for i in range(1, n + 1):
            for j, (vi, dvi) in enumerate(zip(v, dv)):
                h = 1 / i if inverse else i
                vi = h * vi
                if dvi == 0:
                    method = ax.axvline if vertical else ax.axhline
                    args = np.array([vi])
                else:
                    if scale_width:
                        dvi = h * dvi
                    method = ax.axvspan if vertical else ax.axhspan
                    args = np.array([vi - dvi / 2, vi + dvi / 2])
                if inverse:
                    args = sorted(1 / args)
                method(*args, **defaults(kwargs, alpha=1 - np.log(1 + (np.e - 1) * (i - 1) / n)))
                kwargs.pop("label", None)

    @staticmethod
    def add_scale(
        ax,
        scale,
        label=None,
        *,
        vertical=False,
        width=0.01,
        padding=0.1,
        loc="auto",
        color="k",
        fontsize="x-small",
    ):
        """Add a scale patch (a yardstick or ruler)

        Args:
            ax (matplotlib.axes.Axes): The axis to add it to.
            scale (float): The size of the scale in data units.
            label (str | None): A label for the scale.
            vertical (bool): If true, make a vertical one (default is a horizontal one).
            width (float): The line width of the scale in axis units.
            padding (float): The padding between the scale and the axis.
            loc (str): The location of the scale. Can be any of the usual matplotlib locations,
                e.g. ``"auto"``, ``"upper left"``, ``"upper center"``, ``"upper right"``,
                ``"center left"``, ``"center"``, ``"center right"``, ``"lower left"``,
                ``"lower center"`` or ``"lower right"``.
            color (str | tuple): Color for the patch.
            fontsize (str): Font size of the label.

        Returns:
            matplotlib.offsetbox.AnchoredOffsetbox: The artist added
        """
        if loc == "auto":
            loc = "upper left" if vertical else "lower right"
        w, h = scale, width
        w_trans, h_trans = ax.transData, ax.transAxes
        if vertical:  # swap dimensions
            w, h = h, w
            w_trans, h_trans = h_trans, w_trans
        aux = mpl.offsetbox.AuxTransformBox(
            mpl.transforms.blended_transform_factory(w_trans, h_trans)
        )
        aux.add_artist(plt.Rectangle((0, 0), w, h, fc=color))
        if label:
            kwa = dict(text=label, color=color, fontsize=fontsize)
            if vertical:
                aux.add_artist(plt.Text(w * 2, h / 2, ha="left", va="center", rotation=90, **kwa))
            else:
                aux.add_artist(plt.Text(w / 2, h * 1.5, va="bottom", ha="center", **kwa))
        ab = mpl.offsetbox.AnchoredOffsetbox(loc, borderpad=padding, zorder=100, frameon=False)
        ab.set_child(aux)
        ax.add_artist(ab)
        return ab


class XManifoldPlot(XPlot):
    def __init__(self, on_x, on_y, *, on_y_separators=",-+", on_y_subs={}, **kwargs):
        """
        Base class for plotting manifold plots

        A manifold plot consists of multiple subplots, axes and twin axes, all of which
        share the x-axis. The **manifold subplot specification string** ``on_y`` defines what
        is plotted on the y-axes. It should specify a property for each trace, separated by
        ``,`` for each subplot, by ``-`` for twin axes and by ``+`` for traces. For example, the
        string ``"a+b,c-d"`` specifies 2 subplots (a+b and c-d) where on the first subplot the
        traces a and b share the same y-axis and on the second subplot traces c and d have individual y-axis.

        When deriving from this class, you should call :meth:`~.base.XManifoldPlot._create_artists` during init

        Args:
            on_x (str | None): What to plot on the x-axis
            on_y (str or list): What to plot on the y-axis. See :meth:`~.base.XManifoldPlot.parse_nested_list_string`.
                May optionally contain a post-processing function call with the property as first argument
                such as `smooth(key)` or `smooth(key, n=10)`.
            on_y_separators (str): See :meth:`~.base.XManifoldPlot.parse_nested_list_string`
            on_y_subs (dict): See :meth:`~.base.XManifoldPlot.parse_nested_list_string`
            kwargs: Keyword arguments passed to :class:`~.base.XPlot`
        """
        if len(on_y_separators) != 3:
            raise ValueError(
                f"Exactly 3 separators required but got on_y_separators={on_y_separators!r}"
            )

        self.on_x = on_x
        self.on_y, self.on_y_expression = self.parse_nested_list_string(
            on_y, on_y_separators, on_y_subs, strip_off_methods=True
        )
        self._on_y_separators = on_y_separators
        self._on_y_subs = on_y_subs
        self._artists = {}

        super().__init__(
            nrows=len(self.on_y),
            ncols=1,
            nntwins=[len(tw) - 1 for tw in self.on_y],
            sharex="all",
            **kwargs,
        )

        # Format axes
        for i, ppp in enumerate(self.on_y):
            for j, pp in enumerate(ppp):
                a = self.axis(i, j)
                a.set(ylabel=self.label_for(*pp))

                units = np.unique([self.display_unit_for(p) for p in pp])
                if len(units) == 1:
                    if units[0] == "rad":
                        self._set_axis_ticks_angle(a.yaxis, minor=True, deg=False)
                    elif units[0] in ("deg", "°"):
                        self._set_axis_ticks_angle(a.yaxis, minor=True, deg=True)

        self.axis(-1).set(xlabel=self.label_for(self.on_x))

    @property
    def on_y_unique(self):
        return np.unique([p for ppp in self.on_y for pp in ppp for p in pp if p is not None])

    def _create_artists(self, callback, dataset_id=None):
        """Helper method to create artists for subplots and twin axes

        Args:
            callback (callable[int, int, int, matplotlib.axes.Axes, str]): Callback function to create artists.
                Signature: (i, j, k, axis, p) -> artist
                Where i, j, k are the subplot, twin-axis, trace indices respectively;
                axis is the axis and the string p is the property to plot.
            dataset_id (str | None | AUTO): The dataset identifier if this plot represents multiple datasets.
                If :const:`xplt.AUTO`, a new UUID is generated.
        """

        if dataset_id is AUTO:
            dataset_id = str(uuid.uuid4())
        if dataset_id in self._artists:
            raise ValueError(f"Dataset identifier `{dataset_id}` already exists")
        if dataset_id is not None and not isinstance(dataset_id, str):
            raise ValueError(f"The dataset identifier must be a str, but got {type(dataset_id)}")

        self._artists[dataset_id] = []
        for i, ppp in enumerate(self.on_y):
            self._artists[dataset_id].append([])
            for j, pp in enumerate(ppp):
                self._artists[dataset_id][i].append([])
                a = self.axis(i, j)
                for k, p in enumerate(pp):
                    artist = callback(i, j, k, a, p)
                    self._artists[dataset_id][i][j].append(artist)

        self.legend(show="auto")

        return dataset_id

    @property
    def artists(self):
        """Convenient access to artist

        The index can be an (optional) dataset identifier, followed by the subplot, axis and trace index.
        If the dataset identifier is missing, traces are concatenated for all datasets
        Examples: `self.artists[i][j][k]`, `self.artists[i,j,k]`, `self.artists[dataset_id,i,j,k]`
        """

        ALL = object()

        class ArtistIndexHelper:
            """Helper for convenient and backwards-compatible indexing of artists"""

            def __init__(s, dataset=ALL, *indices):
                s.dataset = dataset
                s.indices = indices

            def __iter__(s):
                for i in range(len(s)):
                    yield s[i]

            def __len__(s):
                d = s.dataset if s.dataset is not ALL else list(self._artists)[0]
                if len(s.indices) == 0:
                    return len(self._artists[d])
                elif len(s.indices) == 1:
                    (i,) = s.indices
                    return len(self._artists[d][i])
                elif len(s.indices) == 2:
                    i, j = s.indices
                    if s.dataset is not ALL:
                        return len(self._artists[d][i][j])
                    else:
                        return sum([len(self._artists[d][i][j]) for d in self._artists])

            def __getitem__(s, item):
                if isinstance(item, tuple):
                    return s[item[0]][item[1:]] if len(item) > 1 else s[item[0]]

                elif item is None or isinstance(item, str) or item == ALL:  # select dataset
                    if len(s.indices) > 0:
                        raise IndexError("only the first index may be a string or None")
                    return ArtistIndexHelper(dataset=item)

                elif isinstance(item, int):  # return index for all datasets
                    if len(s.indices) < 2:
                        if item < 0 or item >= len(s):
                            raise IndexError(f"list index {item} out of range 0..{len(s)}")
                        return ArtistIndexHelper(s.dataset, *s.indices, item)
                    else:
                        i, j, k = *s.indices, item
                        if s.dataset is not ALL:
                            return self._artists[s.dataset][i][j][k]
                        else:
                            for d in self._artists:
                                if 0 <= k < len(self._artists[d][i][j]):
                                    return self._artists[d][i][j][k]
                                else:
                                    k -= len(self._artists[d][i][j])
                            raise IndexError(f"list index {item} out of range")

                else:
                    raise IndexError(f"Index must be an int, string, tuple or None. Got {item}.")

            def __setitem__(s, item, value):
                if isinstance(item, tuple):
                    if len(item) > 1:
                        s[item[0]][item[1:]] = value
                    else:
                        s[item[0]] = value

                elif len(s.indices) < 2:
                    raise TypeError("Only single elements may be assigned")

                elif isinstance(item, int):
                    if item < 0 or item >= len(s):
                        raise IndexError(f"list index {item} out of range 0..{len(s)}")

                    i, j, k = *s.indices, item
                    if s.dataset is not ALL:
                        self._artists[s.dataset][i][j][k] = value
                    else:
                        for d in self._artists:
                            if 0 <= k < len(self._artists[d][i][j]):
                                self._artists[d][i][j][k] = value
                                break
                            else:
                                k -= len(self._artists[d][i][j])
                        else:
                            raise IndexError(f"list index {item} out of range")

                else:
                    raise IndexError(f"Index must be an int, string, tuple or None. Got {item}.")

            def __repr__(s):
                info = f"ArtistIndexHelper(dataset={'ALL' if s.dataset is ALL else s.dataset}"
                for i, label in enumerate(("subplot", "axis", "trace")):
                    if len(s.indices) > i:
                        info += f", {label}={s.indices[i]}"
                return info + ")"

        return ArtistIndexHelper()

    def artist(self, name=None, subplot=None, twin=None, trace=None, *, dataset_id=None):
        """Return the artist either by name, or by subplot, twin axes and trace index

        Args:
            name (str | None): Name of the property the artist is plotting
            subplot (int | None): Flat subplot index
            twin (int | None): Twin axis index
            trace (int | None): Trace index
            dataset_id (str | None): The dataset identifier if this plot represents multiple datasets

        Returns:
            matplotlib.artist.Artist: First artist that matches the given criteria
        """
        for i, ppp in enumerate(self.on_y):
            for j, pp in enumerate(ppp):
                for k, p in enumerate(pp):
                    if (
                        (p == name or name is None)
                        and (i == subplot or subplot is None)
                        and (j == twin or twin is None)
                        and (k == trace or trace is None)
                    ):
                        return self.artists[dataset_id][i][j][k]

    def axis(self, subplot_or_name=0, twin=0, *, subplot=None, name=None):
        """Return the axis for a given flat subplot index and twin index, or an artist name

        Either get axis by index ``axis(subplot, twin)`` or by property name ``axis(name)``.

        Args:
            subplot_or_name (int | str): Flat subplot index or property name
            subplot (int): Flat subplot index
            twin (int): Twin index
            name (str | None): Name of the property the axis is plotting. If this is not None, subplot and twin are ignored.

        Returns:
            matplotlib.axes.Axes: Axis for the given subplot and twin index
        """
        if isinstance(subplot_or_name, str):
            name = subplot_or_name
        elif subplot is None:
            subplot = subplot_or_name
        if name is not None:
            if subs := self._on_y_subs.get(name):
                for n in subs.split(self._on_y_separators[2]):
                    try:
                        return self.axis(name=n)
                    except ValueError:
                        continue
            else:
                for i, ppp in enumerate(self.on_y):
                    for j, pp in enumerate(ppp):
                        for k, p in enumerate(pp):
                            if p == name:
                                return self.axis(i, j)
            raise ValueError(f"Property {name} not found")
        return super().axis(subplot, twin)

    def _legend_label_for(self, p):
        """
        Return legend label for a single property

        Args:
            p (str | tuple): Property name or tuple of (subplot, twin, trace) index

        Returns:
            str: Legend label
        """
        if isinstance(p, str):
            return super()._legend_label_for(p)
        else:
            i, j, k = p
            label = super()._legend_label_for(self.on_y[i][j][k])
            if w := self.on_y_expression[i][j][k]:
                w = re.split(r"\b", w)[1]
                label += f" ({w})"
            return label

    def legend(self, subplot="all", show=True, **kwargs):
        """Add, update or remove legend for a subplot

        Args:
            subplot (int | list | str): Subplot axis index or indices
            show (bool | str): If True, show the legend. If "auto", show
                legend for subplots with more than one trace, or if subplot is specified explicitly.
            kwargs: Keyword arguments passed to :meth:`matplotlib.axes.Axes.legend`

        """
        if subplot == "all":
            subplot = range(len(self.axflat))
        if isinstance(subplot, int):
            subplot = [subplot]
            if show == "auto":
                show = True  # always show legend if single subplot is specified

        set_kwargs = {k: kwargs.pop(k) for k in ("in_layout",) if k in kwargs}

        for s in subplot:
            # aggregate handles and use topmost axes for legend
            handles = []
            for ax in self.axes(s):
                handles.extend(ax.get_legend_handles_labels()[0])
            if not show or (show == "auto" and len(handles) <= 1):
                if ax.get_legend():
                    ax.get_legend().remove()
            else:
                # join handles
                handle_map = {
                    h: [h] for h in handles if not hasattr(h, "_join_legend_entry_with")
                }
                for h in handles:
                    if main_handle := getattr(h, "_join_legend_entry_with", None):
                        handle_map[main_handle].append(h)
                handles = [tuple(hs) for hs in handle_map.values()]
                labels = [h.get_label() for h in handle_map]

                # show legend
                if len(handles) > 0:
                    legend = ax.legend(handles=handles, labels=labels, **kwargs)
                    legend.set(**set_kwargs)

    def autoscale(self, subplot="all", *, axis="xy", reset=False, freeze=True, tight=None):
        """Autoscale the axes of a subplot

        Args:
            subplot (int | iterable | str): Subplot axis index, indices or "all"
            axis (str | None | bool): Whether and on which axes to perform autoscaling.
                One of `"x"`, `"y"`, `"xy"`, `False` or `None`. If `None`, decide based on :meth:`matplotlib.axes.Axes.get_autoscalex_on` and :meth:`matplotlib.axes.Axes.get_autoscaley_on`.
                For backwards compatibility, the following aliases are also supported: `"both"`, `True`, `""`.
            reset (bool): Whether to ignore any data limits already registered.
            freeze (bool): Whether to keep the updated axes limits (True) or enable automatic
                autoscaling on future draws (for all present and new artists).
            tight (str | None): Enables tight scaling without margins for the specified dimension.
                May be ``"x"``, ``"y"``, ``"both"`` or ``None``.
        """
        kwargs = dict(axis=axis, reset=reset, freeze=freeze, tight=tight)

        if subplot == "all":
            subplot = range(len(self.axflat))

        for s in flattened(subplot):
            self._autoscale(self.axflat[s], **kwargs)
            for i, axt in enumerate(self.axflat_twin[s]):
                self._autoscale(axt, **kwargs)

    def axline(self, kind, val, **kwargs):
        """Plot a vertical or horizontal line for a given coordinate

        Args:
            kind (str): property at which to place the line (e.g. "s", "x", "betx", etc.)
            val (float): Value of property.
            kwargs: See :meth:`~.base.XManifoldPlot.axspan`.

        """
        self.axspan(kind, val, None, **kwargs)

    def axspan(
        self,
        kind,
        val,
        val_to=None,
        *,
        subplots="all",
        annotation=None,
        annotation_loc=None,
        annotation_kwargs=None,
        **kwargs,
    ):
        """Plot a vertical or horizontal span (or line) for a given coordinate

        Args:
            kind (str): property at which to place the line (e.g. "s", "x", "betx", etc.).
            val (float): Value of property.
            val_to (float | None): Second value of property to plot a span. If this is `None`, plot a line instead of a span.
            subplots (list[int]): Subplots to plot line onto. Defaults to all with matching coordinates.
            annotation (string | None): Optional text annotation for the line or span. Use this to place
                text on the axes. To put text in the legend, use `label=...`.
            annotation_loc (float, string): Location of annotation along line or span as fraction between 0 and 1,
                or a string ``"lower"``, ``"bottom"``, ``left"``, ``"upper"``, ``"top"``, ``"right"``, ``center``.
                For vertical lines or spans extending over multiple subplots, 0 is the bottom of the lowermost
                and 1 the top of the uppermost subplot.
            annotation_kwargs (dict | None): Arguments for :meth:`matplotlib.axes.Axes.text`.
            kwargs: Arguments passed to :meth:`matplotlib.axes.Axes.axvspan` or :meth:`matplotlib.axes.Axes.axhspan`
                (or :meth:`matplotlib.axes.Axes.axvline` or :meth:`matplotlib.axes.Axes.axhline` if `val_to` is `None`)

        """
        if subplots == "all":
            subplots = list(range(len(self.on_y)))

        if isinstance(annotation_loc, str):
            if annotation_loc in ("lower", "bottom", "left"):
                annotation_loc = 0
            elif annotation_loc in ("upper", "top", "right"):
                annotation_loc = 1
            elif annotation_loc in ("center", "centre"):
                annotation_loc = 0.5
            else:
                raise ValueError(f"Invalid annotation location: {annotation_loc}")

        if val_to is None:  # only a line
            kwargs = defaults_for("plot", kwargs, color="k", lw=1, zorder=1.9)
        else:  # a span
            kwargs = defaults_for(
                "fill_between", kwargs, color="lightgray", zorder=1.9, lw=0, alpha=0.6
            )

        val = val * self.factor_for(kind)
        if val_to is not None:
            val_to = val_to * self.factor_for(kind)

        def plot_line_with_annotation(hor, loc, with_annotation=True):
            if val_to is None:
                line = (a.axhline if hor else a.axvline)(val, **kwargs)
                color = line.get_color()
            else:
                line = (a.axhspan if hor else a.axvspan)(val, val_to, **kwargs)
                color = line.get_facecolor()
            if annotation and with_annotation:
                align = int(np.clip(round(2 * loc), 0, 2))  # 0, 1 or 2
                text_kwargs = defaults_for(
                    "text",
                    annotation_kwargs,
                    fontsize="xx-small",
                    c=color,  # RGBA tuple
                    alpha=1,  # overwrites color[3]
                    zorder=line.get_zorder(),
                    ha=["left", "center", "right"][align] if hor else "right",
                    va="bottom" if hor else ["bottom", "center", "top"][align],
                    rotation=0 if hor else 90,
                )
                a.text(
                    *(val, loc)[:: -1 if hor else 1],
                    s=f" {annotation} ",
                    transform=mpl.transforms.blended_transform_factory(
                        *(a.transData, a.transAxes)[:: -1 if hor else 1]
                    ),
                    **text_kwargs,
                )

        for i, ppp in enumerate(self.on_y):
            if i not in subplots:
                continue

            for j, pp in enumerate(ppp):
                a = self.axis(i, j)
                if kind == self.on_x:  # vertical line or span
                    loc = annotation_loc if annotation_loc is not None else 1
                    # rescale from global to subplot coordinates:
                    loc = (max(subplots) - min(subplots) + 1) * loc - (max(subplots) - i)
                    plot_line_with_annotation(hor=False, loc=loc, with_annotation=0 <= loc <= 1)
                    break  # skip twin axes

                else:
                    # horizontal line or span
                    for k, p in enumerate(pp):
                        if p == kind:  # axis found
                            loc = annotation_loc if annotation_loc is not None else 0
                            plot_line_with_annotation(hor=True, loc=loc)

    @staticmethod
    def parse_nested_list_string(
        list_string, separators=",-+", subs={}, *, strip_off_methods=False
    ):
        """Parse a separated string or nested list or a mixture of both




        Args:
            list_string (str or list): The string or nested list or a mixture of both to parse.
            separators (str): The characters that separate the elements. The number of characters
                determines the depth of the returned list.
            subs (dict): A dictionary of substitutions to apply to the elements during parsing.
                May introduce additional separators of equal or deeper level.
            strip_off_methods (bool): If true, each element can be a name `name` or an expression
                in the form `method(name, ...)`. The methods are stripped off, and returned separately.

        Returns:
            nested list of names in the string,
            nested list of expressions in the string (only if strip_off_methods is True)

        Example:
            >>> XManifoldPlot.parse_nested_list_string("a+b, c-d,fun(e,2)")
            [[['a', 'b']], [['c'], ['d']], [['fun(e,2)']]]
            >>> XManifoldPlot.parse_nested_list_string("a+b, c-d,fun(e,2)", strip_off_methods=True)
            ([[['a', 'b']], [['c'], ['d']], [['e']]],
             [[[None, None]], [[None], [None]], [['fun(e,2)']]])
        """

        def savesplit(string, sep):
            """Split the string at sep except inside parantheses"""
            return re.split(f"\\s*\\{sep}\\s*(?![^()]*\\))", string)

        if type(list_string) is str:
            elements = []
            for element in savesplit(list_string.strip(), separators[0]):
                element = subs.get(element, element)
                # split again in case subs contains a separator
                elements.extend(savesplit(element, separators[0]))
        else:
            elements = list(list_string)
        expressions = [None] * len(elements)
        if len(separators) > 1:
            for i in range(len(elements)):
                result = XManifoldPlot.parse_nested_list_string(
                    elements[i], separators[1:], subs, strip_off_methods=strip_off_methods
                )
                if strip_off_methods:
                    elements[i], expressions[i] = result
                else:
                    elements[i] = result
        elif strip_off_methods:
            for i in range(len(elements)):
                if isinstance(elements[i], str):
                    if m := re.match(r".*\(([^(),\s]+)[^()]*\)", elements[i]):
                        name, expression = m.groups()[0], m.string
                        elements[i] = name
                        expressions[i] = expression
        return (elements, expressions) if strip_off_methods else elements


## Restrict star imports to local namespace
__all__ = [
    name
    for name, thing in globals().items()
    if not (name.startswith("_") or isinstance(thing, types.ModuleType))
]



import pytest
import pytest_benchmark
import numpy as np
from numpy.testing import assert_equal
import xplt.util


@pytest.fixture(scope="function")
def benchmark_ref(request):
    # See https://github.com/ionelmc/pytest-benchmark/issues/166
    bm = pytest_benchmark.plugin.benchmark.__pytest_wrapped__.obj(request)
    return next(bm) if hasattr(bm, "__next__") else bm


@pytest.mark.parametrize(
    "ndata,nbins,exp_time_fract",
    [
        (1_000, 2**17, 1 / 2),
        (1_000, 2**20, 1 / 4),
        (1_000_000, 2**20, 1 / 3),
        (1_000_000, 2**23, 1 / 10),
        (1_000_000, 2**27, 1 / 32),
    ],
)
def test_binned_data(benchmark, benchmark_ref, ndata, nbins, exp_time_fract):
    data = np.random.uniform(0, 1, ndata)

    v_min, dv, c_bin = benchmark(xplt.util.binned_data, data, n=nbins, v_range=(0, 1))
    c_ref = benchmark_ref(np.histogram, data, bins=nbins, range=(0, 1))[0]

    # test result
    assert v_min == 0
    assert dv == 1 / nbins
    assert np.sum(c_bin) == ndata
    assert c_bin.size == nbins
    assert_equal(c_bin, c_ref)

    # test performance
    fract = benchmark.stats.stats.median / benchmark_ref.stats.stats.median
    assert fract < exp_time_fract, (
        f"Performance of binned_data compared to np.histogram is {fract:g} but expected"
        f" {exp_time_fract} or less ({benchmark.stats.stats.median:g}s vs {benchmark_ref.stats.stats.median:g}s"
    )


def test_averaging():

    data = np.array([1, 2, 4, 5, 6, 8, 3, 10, 25, 13])

    avg = xplt.util.average(data, n=2)
    assert_equal(avg, [1.5, 4.5, 7, 6.5, 19])
    avg = xplt.util.average(data, n=2, keepdim=True)
    assert_equal(avg, [1.5, 1.5, 4.5, 4.5, 7, 7, 6.5, 6.5, 19, 19])
    avg = xplt.util.average(data, n=3, function=np.max)
    assert_equal(avg, [4, 8, 25])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Methods for plotting phase space distributions"""

__author__ = "Philipp Niedermayer"
__contact__ = "eltos@outlook.de"
__date__ = "2022-09-06"


from matplotlib.patches import Ellipse
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from .util import *
from .base import XPlot, XManifoldPlot, AngleLocator, RadiansFormatter
from .particles import ParticlePlotMixin

pairwise = np.c_


PUBLIC_SECTION_BEGIN()


class PhaseSpacePlot(XPlot, ParticlePlotMixin):
    """A plot for phase space distributions"""

    def __init__(
        self,
        particles=None,
        kind=None,
        plot="auto",
        *,
        scatter_kwargs=None,
        hist_kwargs=None,
        mask=None,
        masks=None,
        color=None,
        cmap="magma_r" or "Blues",
        cbar_loc="auto",
        projections="auto",
        projections_kwargs=None,
        mean=False,
        mean_kwargs=None,
        std=False,
        std_kwargs=None,
        percentiles=None,
        percentile_kwargs=None,
        nrows=None,
        ncols=None,
        titles="auto",
        animated=False,
        twiss=None,
        **kwargs,
    ):
        """

        Args:
            particles (Any): A dictionary with particle information
            kind (str | list): Defines the properties to plot.
                This can be a nested list or a separated string or a mixture of lists and strings where
                the first list level (or separator ``,``) determines the subplots,
                and the second list level (or separator ``-``) determines coordinate pairs.
                In addition, abbreviations for x-y-parameter pairs are supported (e.g. 'x' for 'x-px').
                For normalized coordinates, use uppercase letters (e.g. 'X' for 'X-Px').
            plot (str): Defines the type of plot. Can be ``"auto"``, ``"scatter"`` or ``"hist"``. Default is ``"auto"``
                for which the plot type is chosen automatically based on the number of particles.
            scatter_kwargs (dict): Additional kwargs for scatter plot, see :meth:`matplotlib.axes.Axes.scatter`.
            hist_kwargs (dist): Additional kwargs for 2D histogram plot, see :meth:`matplotlib.axes.Axes.hexbin`.
            mask (Any): An index mask to select particles to plot. If None, all particles are plotted.
            masks (list[mask]): List of masks for each subplot.
            color (str | list[str]): Properties defining the color of points for the scatter plot(s). Implies plot='scatter'.
                Pass a list of properties to use different values for each subplot
            cmap (str): Colormap to use for the hist plot.
            cbar_loc (str): Location of the colorbar, such as 'auto', 'right', 'inside upper right', etc.
                Use None to disable colorbar.
            projections (bool | str | list): Add histogrammed projections onto axis. Can be ``True``, ``False``, ``"x"``,
                ``"y"``, ``"auto"`` or a list of these for each subplot.
            projections_kwargs (dict): Additional kwargs for histogram projection, see :meth:`matplotlib.axes.Axes.step`.
            mean (bool | list): Whether to indicate mean of distribution with a cross marker. Boolean or list of booleans for each subplot.
            mean_kwargs (dict): Additional kwargs for marker, see :meth:`matplotlib.axes.Axes.plot`.
            std (bool | list): Whether to indicate standard deviation of distribution with an ellipse. Boolean or list of booleans for each subplot.
            std_kwargs (dict): Additional kwargs for std ellipses, see :class:`matplotlib.patches.Ellipse`.
            percentiles (list): List of percentiles (in percent) to indicate in the distribution with ellipses.
                Can also be a list of lists for each subplot.
            percentile_kwargs (dict): Additional kwargs for percentile ellipses, see :class:`matplotlib.patches.Ellipse`.
            ncols (int | None): Number of columns in subplot layout. If None, the layout is determined automatically.
            nrows (int | None): Number of columns in subplot layout. If None, the layout is determined automatically.
            titles (list[str]): List of titles for each subplot or 'auto' to automatically set titles based on plot kind.
            animated (bool): If True, improve plotting performance for creating an animation.
            twiss (dict | None): Twiss parameters (alfx, alfy, betx and bety) to use for conversion to normalized phase space coordinates.
            kwargs: See :class:`~.particles.ParticlePlotMixin` and :class:`~.base.XPlot` for additional arguments


        """

        # parse kind string by splitting at commas and dashes and replacing abbreviations
        abbreviations = dict(
            x="x-px",
            y="y-py",
            z="zeta-delta",
            X="X-Px",
            Y="Y-Py",
        )
        if kind is None:
            kind = "x,y,z" if twiss is None else "X,Y,z"
        self.kind = XManifoldPlot.parse_nested_list_string(
            XManifoldPlot.parse_nested_list_string(kind, ",", abbreviations),
            ",-",  # no abbreviations on second level!
        )
        if np.any([len(k) != 2 for k in self.kind]):
            raise ValueError("Kind must only contain exactly two coordinates per subplot")
        n = len(self.kind)

        # sanitize parameters
        if not hasattr(mean, "__iter__"):
            mean = n * [mean]
        if not hasattr(std, "__iter__"):
            std = n * [std]
        if percentiles is None or not hasattr(percentiles[0], "__iter__"):
            percentiles = n * [percentiles]
        if isinstance(projections, str) or not hasattr(projections, "__iter__"):
            projections = n * [projections]

        if len(mean) != n:
            raise ValueError(f"mean must be a boolean or a list of length {n}, but got {mean!r}")
        if len(std) != n:
            raise ValueError(f"std must be a boolean or a list of length {n}, but got {std!r}")
        if len(percentiles) != n:
            raise ValueError(
                f"percentiles list must be flat or of length {n}, but got {percentiles!r}"
            )
        if plot not in ["auto", "scatter", "hist"]:
            raise ValueError(f"plot must be 'auto', 'scatter' or 'hist', but got {plot!r}")
        if color is not None:
            if plot not in ["auto", "scatter"]:
                raise ValueError(f"Setting color requires plot='scatter', but got plot={plot!r}")
            plot = "scatter"
            if isinstance(color, str):
                color = np.resize(color.split(","), n)
            if cbar_loc == "auto":
                cbar_loc = (
                    "right"
                    if len(color) <= 1 or np.all(color == color[0])
                    else "inside upper right"
                )
        else:
            color = n * [None]

        self.plot = plot
        self.color = color
        self.percentiles = percentiles
        self.projections = projections

        # Create plot axes

        # initialize figure with n subplots
        if ncols is None:
            if nrows is None:
                nrows = int(np.sqrt(n))
                while n % nrows != 0:
                    nrows -= 1
            ncols = n // nrows
        if ncols * nrows != n:
            raise ValueError(
                f"Layout with {ncols} columns and {nrows} rows conflicts with {n} plots!"
            )
        kwargs = defaults(kwargs, figsize=(4 * ncols, 4 * nrows))

        kwargs = self._init_particle_mixin(twiss=twiss, **kwargs)
        super().__init__(nrows=nrows, ncols=ncols, **kwargs)

        if len(self.axflat) < n:
            raise ValueError(f"Need {n} axes but got only {len(self.axflat)}")

        # Create distribution plots
        self.artists_scatter = [None] * n
        self.artists_hexbin = [()] * n
        self.artists_mean = [None] * n
        self.artists_std = [None] * n
        self.artists_percentiles = [()] * n
        self.ax_twin = [{} for _ in range(n)]
        self.artists_twin = [{} for _ in range(n)]
        self.artists_hamiltonian = [{} for _ in range(n)]

        for i, ((a, b), c, ax) in enumerate(zip(self.kind, self.color, self.axflat)):
            # we handle autoscaling manually
            ax.set_autoscale_on(False)

            # 2D phase space distribution
            ##############################

            # scatter plot
            kwargs = defaults_for(
                "scatter", scatter_kwargs, s=4, cmap=cmap, lw=0, animated=animated
            )
            scatter_cmap = kwargs.pop("cmap")  # bypass UserWarning: ignored
            vmin, vmax = kwargs.pop("vmin", None), kwargs.pop("vmax", None)
            self.artists_scatter[i] = ax.scatter([], [], **kwargs)
            # TODO: if color is None, use ax.plot([], [], marker=".", ls="") since it is faster
            self.artists_scatter[i].cmap = (
                mpl.colormaps[scatter_cmap] if isinstance(scatter_cmap, str) else scatter_cmap
            )
            self.artists_scatter[i].vmin_vmax = vmin, vmax
            # add colorbar
            if c is not None and (np.any(self.color != c) or i == n - 1):
                cbargs = dict(label=self.label_for(c))
                if self.display_unit_for(c) == "rad":
                    cbargs.update(ticks=AngleLocator(deg=False), format=RadiansFormatter())
                if cbar_loc and cbar_loc.startswith("inside "):
                    # colorbar inside plot
                    axins = inset_axes(
                        ax, width="50%", height="3%", loc=cbar_loc.lstrip("inside ")
                    )
                    self.fig.colorbar(
                        self.artists_scatter[i], **cbargs, cax=axins, orientation="horizontal"
                    )
                    axins.tick_params(labelsize=8)
                    axins.xaxis.offsetText.set_fontsize(8)
                    axins.xaxis.label.set_size(8)
                elif cbar_loc:
                    self.fig.colorbar(self.artists_scatter[i], **cbargs, ax=ax, location=cbar_loc)

            # hexbin histogram
            self._hxkw = defaults_for(
                "hexbin", hist_kwargs, mincnt=1, cmap=cmap, rasterized=True, animated=animated
            )

            # 2D mean indicator
            if mean[i]:
                kwargs = defaults_for(
                    "plot", mean_kwargs, color="k", marker="+", ms=8, zorder=100
                )
                (self.artists_mean[i],) = ax.plot([], [], **kwargs, animated=animated)

            # 2D std ellipses
            if std[i]:
                kwargs = defaults_for(
                    mpl.patches.Ellipse,
                    std_kwargs,
                    fill=False,
                    color="k",
                    lw=1,
                    ls="-",
                    zorder=100,
                )
                self.artists_std[i] = Ellipse((0, 0), 0, 0, **kwargs, animated=animated)
                ax.add_artist(self.artists_std[i])

            # 2D percentile ellipses
            if percentiles[i]:
                self.artists_percentiles[i] = []
                for j, _ in enumerate(self.percentiles[i]):
                    kwargs = defaults_for(
                        mpl.patches.Ellipse,
                        percentile_kwargs,
                        fill=False,
                        color="k",
                        lw=1,
                        ls=(0, [5, 5] + [1, 5] * j),
                        zorder=100,
                    )
                    artist = Ellipse((0, 0), 0, 0, **kwargs, animated=animated)
                    ax.add_artist(artist)
                    self.artists_percentiles[i].append(artist)

            # Axes title and labels
            if titles == "auto":
                title = self.title_for(a, b)
            elif titles is None:
                title = None
            else:
                title = titles[i]
            ax.set(title=title, xlabel=self.label_for(a), ylabel=self.label_for(b))

            # Set angle ticks
            for ab, axis in ((a, ax.xaxis), (b, ax.yaxis)):
                if self.display_unit_for(ab) == "rad":
                    self._set_axis_ticks_angle(axis, minor=True, deg=False)
                elif self.display_unit_for(a) in ("deg", "°"):
                    self._set_axis_ticks_angle(axis, minor=True, deg=True)

            # 1D histogram projections
            ###########################

            if self.projections[i]:
                kwargs = defaults_for("plot", projections_kwargs, color="k", alpha=0.3, lw=1)
                for xy, yx in zip("xy", "yx"):
                    if self.projections[i] != yx:
                        # Create twin xy axis and artists
                        twin = ax.twinx() if xy == "x" else ax.twiny()
                        twin.set(**{f"{yx}ticks": [], f"{yx}lim": (0, 0.2)})
                        self.ax_twin[i][xy] = twin
                        (hist,) = twin.step([], [], **kwargs, animated=animated)
                        self.artists_twin[i][xy] = hist

        # set data
        if particles is not None:
            self.update(particles, mask=mask, masks=masks, autoscale=True)

    def update(self, particles, *, mask=None, masks=None, autoscale=False):
        """
        Update the data this plot shows

        Args:
            particles (Any): A dictionary with particle information
            mask (Any): An index mask to select particles to plot. If None, all particles are plotted.
            masks (list): List of masks for each subplot.
            autoscale (bool): Whether or not to perform autoscaling on all axes.

        Returns:
            List of changed artists.
        """
        if masks is None:
            masks = [mask] * len(self.kind)
        elif mask is not None:
            raise ValueError("Only one of mask and masks can be set.")
        if len(masks) != len(self.kind):
            raise ValueError(f"masks must be a list of length {len(self.kind)}")

        changed_artists = []

        for i, ((a, b), c, ax) in enumerate(zip(self.kind, self.color, self.axflat)):
            # coordinates
            x = self.prop(a).values(particles, masks[i], unit=self.display_unit_for(a))
            y = self.prop(b).values(particles, masks[i], unit=self.display_unit_for(b))

            # statistics
            XY = np.array((x, y))

            XY0 = np.mean(XY, axis=1)
            UV = XY - XY0[:, np.newaxis]  # centered coordinates

            # 2D phase space distribution
            ##############################

            plot = self.plot
            if plot == "auto":
                plot = "scatter" if len(x) <= 1000 else "hist"

            # scatter plot
            scatter = self.artists_scatter[i]
            if plot == "scatter":
                scatter.set_visible(True)
                scatter.set_offsets(pairwise[x, y])
                if c is not None:
                    v = self.prop(c).values(particles, masks[i], unit=self.display_unit_for(c))
                    if autoscale:
                        # scatter.set_clim(np.min(v), np.max(v)) # sometimes leaves behind black dots (bug?)
                        # scatter.norm = mpl.colors.Normalize(np.min(v), np.max(v)) # works, but resets colorbar locator/formatter
                        # scatter.norm.autoscale(v) # also sometimes leaves behind black dots (bug?)

                        if scatter.colorbar is not None:
                            locator, formatter = (
                                scatter.colorbar.locator,
                                scatter.colorbar.formatter,
                            )
                        vmin, vmax = scatter.vmin_vmax
                        scatter.norm = mpl.colors.Normalize(vmin or np.min(v), vmax or np.max(v))
                        if scatter.colorbar is not None:
                            scatter.colorbar.locator, scatter.colorbar.formatter = (
                                locator,
                                formatter,
                            )

                    cm = mpl.cm.ScalarMappable(norm=scatter.norm, cmap=scatter.cmap)
                    scatter.set_facecolor(cm.to_rgba(v))

                changed_artists.append(scatter)
            elif scatter.get_visible():
                scatter.set_visible(False)
                changed_artists.append(scatter)

            # hexbin plot
            # remove old hexbin and create a new one (no update method exists)
            for artist in self.artists_hexbin[i]:
                changed_artists.append(artist)
                artist.remove()
            self.artists_hexbin[i] = []
            if plot == "hist":
                # twice to mitigate https://stackoverflow.com/q/17354095
                hexbin_bg = ax.hexbin(x, y, edgecolors="face", lw=0.1, **self._hxkw)
                hexbin_fg = ax.hexbin(x, y, edgecolors="none", lw=0.0, **self._hxkw)
                self.artists_hexbin[i] = [hexbin_bg, hexbin_fg]
                changed_artists.extend(self.artists_hexbin[i])

            # 2D mean indicator (cross)
            if self.artists_mean[i]:
                self.artists_mean[i].set_data(XY0.reshape([2, 1]))
                changed_artists.append(self.artists_mean[i])

            # 2D size indicator (ellipses)
            if UV.shape[1] > 1 and (self.artists_std[i] or self.artists_percentiles[i]):
                evals, evecs = np.linalg.eig(np.cov(UV))  # eigenvalues and -vectors

                # 2D std indicator
                if self.artists_std[i]:
                    w, h = 2 * np.sqrt(evals)
                    self.artists_std[i].set(
                        center=XY0, width=w, height=h, angle=np.degrees(np.arctan2(*evecs[1]))
                    )
                    changed_artists.append(self.artists_std[i])

                # 2D percentile indicator
                if self.artists_percentiles[i]:
                    # normalize distribution using eigenvalues and -vectors
                    NN = np.dot(evecs.T, UV) / np.sqrt(evals)[:, np.newaxis]
                    for j, p in enumerate(self.percentiles[i]):
                        # percentile in normalized distribution
                        e = np.percentile(np.sum(NN**2, axis=0), p) ** 0.5
                        w, h = 2 * e * np.sqrt(evals)
                        self.artists_percentiles[i][j].set(
                            center=XY0, width=w, height=h, angle=np.degrees(np.arctan2(*evecs[1]))
                        )
                        changed_artists.append(self.artists_percentiles[i][j])

            # Autoscale
            if autoscale:
                if plot == "scatter":
                    self._autoscale(ax, artists=[self.artists_scatter[i]])
                elif plot == "hist":
                    self._autoscale(ax, artists=self.artists_hexbin[i])

            # 1D histogram projections
            ###########################

            project = self.projections[i]
            if project == "auto":
                project = len(x) >= 200

            for xy, v in zip("xy", (x, y)):
                if xy in self.artists_twin[i]:
                    hist = self.artists_twin[i][xy]
                    hist.set_visible(project)
                    if project:
                        # 1D histogram
                        rng = getattr(ax, f"get_{xy}lim")()
                        counts, edges = np.histogram(v, bins=101, range=rng)
                        counts = counts / len(v)

                        steps = (np.append(edges, edges[-1]), np.concatenate(([0], counts, [0])))
                        hist.set_data(steps if xy == "x" else steps[::-1])
                        changed_artists.append(hist)

        return changed_artists

    def title_for(self, a, b):
        """
        Plot title for a given pair (a,b) of properties
        """
        titles = {  # Order of a,b does not matter
            "x-px": "Horizontal phase space",
            "y-py": "Vertical phase space",
            "zeta-delta": "Longitudinal phase space",
            "zeta_wrapped-delta": "Longitudinal phase space",
            "X-Px": "Normalized horizontal phase space",
            "Y-Py": "Normalized vertical phase space",
            "x-y": "Transverse profile",
        }
        title = titles.get(f"{a}-{b}", titles.get(f"{b}-{a}"))
        if title is None:
            style = dict(unit=False, description=False)
            title = f"{self.label_for(a, **style)}-{self.label_for(b, **style)} phase space"
        return title

    def axline(self, kind, val, *, subplots="all", also_on_normalized=False, delta=0, **kwargs):
        """Plot a vertical or horizontal line for a given coordinate

        Args:
            kind (str): Phase space coordinate
            val (float): Value of phase space coordinate
            subplots (list of int): Subplots to plot line onto. Defaults to all with matching coordinates.
            also_on_normalized (bool): If true, also plot line for related (de-)normalized phase space coordinates.
            delta (float): The momentrum error used to convert to (de-)normalized  phase space coordinates.
            kwargs: Arguments for axvline or axhline

        """

        kwargs = defaults_for("plot", kwargs, color="k")

        for i, (ab, ax) in enumerate(zip(self.kind, self.axflat)):
            if subplots != "all" and i not in subplots:
                continue

            for a, line in zip(ab, (ax.axvline, ax.axhline)):
                if a == kind:
                    # same axis found, draw line
                    line(val * self.factor_for(kind), **kwargs)

                elif also_on_normalized and self.twiss is not None and kind.lower() == a.lower():
                    # related (de-)normalized axis found, transform and draw line
                    xy = kind.lower()[-1]
                    if kind in "x,y":  # x,y -> X,Y
                        v = normalized_coordinates(val, 0, self.twiss, xy, delta)[0]
                    elif kind in "X,Y":  # X,Y -> x,y
                        v = denormalized_coordinates(val, 0, self.twiss, xy, delta)[0]
                    elif kind in "px,py":  # px,py -> Px,Py
                        v = normalized_coordinates(0, val, self.twiss, xy, delta)[1]
                    elif kind in "Px,Py":  # Px,Py -> px,py
                        v = denormalized_coordinates(0, val, self.twiss, xy, delta)[1]
                    else:
                        continue
                    line(v * self.factor_for(a), **kwargs)

    def plot_hamiltonian_kobayashi(
        self,
        subplots,
        S,
        mu,
        *,
        delta=0,
        extend=1,
        autoscale=1.1,
        separatrix=True,
        separatrix_kwargs=None,
        equipotentials=True,
        equipotentials_kwargs=None,
    ):
        """Plot separatrix and equipotential lines of kobayashi hamiltonian

        Args:
            subplots (int or list of int): Index of subplot(s)
            S (float): Virtual sextupole strength in m^(-1/2)
            mu (float): Virtual sextupole phase in rad/2pi
            delta (float): Momentum offset.
            extend (float): Extend of separatrix and equipotential lines. If > 1 they are drawn beyond the the stable region.
            autoscale (bool | float): Whether to autoscale axis or not (bool), or the extend to consider for autoscaling
            separatrix (bool): Plot separatrix. Defaults to True.
            separatrix_kwargs (dict | None): Keyword arguments for separatrix line plot.
            equipotentials (bool): Plot equipotential lines. Defaults to True.
            equipotentials_kwargs (dict | None): Keyword arguments for equipotential line contour plot.
        """
        if not hasattr(subplots, "__iter__"):
            subplots = [subplots]

        for subplot in subplots:
            ax = self.axflat[subplot]
            a, b = self.kind[subplot]
            config = {
                "x-px": ("x", False),
                "y-py": ("y", False),
                "X-Px": ("x", True),
                "Y-Py": ("y", True),
            }
            if f"{a}-{b}" in config:
                xy, normalized = config[f"{a}-{b}"]
                swap = False
            elif f"{b}-{a}" in config:
                xy, normalized = config[f"{b}-{a}"]
                swap = True
            else:
                raise TypeError(f"Invalid plot kind {a}-{b} for kobayashi hamiltonian")

            # twiss parameters at plot location
            if self.twiss is None:
                raise ValueError("No twiss parameters provided during plot creation")
            elif not np.isscalar(self.twiss["betx"]) and len(self.twiss["betx"]) != 1:
                raise ValueError(
                    f"Twiss table has {len(self.twiss)} entries, expected exactly 1. "
                    "Did you forget to specify at_elements with a single element during twiss?"
                )
            alfx, betx, mux, x, px = [
                get(self.twiss, pre + xy) for pre in ["alf", "bet", "mu", "", "p"]
            ]
            q = get(self.twiss, "q" + xy)
            q = q + delta * get(self.twiss, "dq" + xy)

            # distance d to nearby 3rd order resonance r
            r = round(3 * q) / 3
            d = q - r
            # size of stable triangle
            h = 4 * np.pi * d / S
            # phase advance from virtual sextupole to here
            dmu = 2 * np.pi * (mux - mu)
            rotation = np.array([[+np.cos(dmu), np.sin(dmu)], [-np.sin(dmu), np.cos(dmu)]])

            def transform(XY):
                """Transform normalized phase space coordinates into coordinates of plot"""
                # rotate so as to account for phase advance
                XY = np.tensordot(rotation, XY, 1)
                # match plot settings
                if not normalized:
                    XY = np.array(denormalized_coordinates(*XY, self.twiss, xy, delta))
                if swap:
                    XY = XY[::-1]
                XY[0] *= self.factor_for(a)
                XY[1] *= self.factor_for(b)
                return XY

            # plot separatrix
            if separatrix:
                kwarg = defaults_for(
                    "plot", separatrix_kwargs, color="r", ls="--", label="Separatrix"
                )

                t = extend
                for X, Y in (  # triangle edges
                    [(-2, -2), (-2 * t, 2 * t)],
                    [(1 - 3 * t, 1 + 3 * t), (1 + t, 1 - t)],
                    [(1 - 3 * t, 1 + 3 * t), (-1 - t, -1 + t)],
                ):
                    X = h * np.array(X) / 2
                    Y = h * 3**0.5 * np.array(Y) / 2
                    ax.plot(*transform((X, Y)), **kwarg)
                    kwarg.pop("label", None)

            # plot equipotential lines
            if equipotentials:
                kwargs = defaults(
                    equipotentials_kwargs, colors="lightgray", linewidths=1, alpha=0.5
                )

                # Hamiltonian normalized to value at separatrix
                X = extend * h * np.linspace(-1, 2, 500)
                Y = extend * h * 3**0.5 * np.linspace(-1, 1, 500)
                X, Y = np.meshgrid(X, Y)
                H = (3 * h * (X**2 + Y**2) + 3 * X * Y**2 - X**3) / h**3 / 4
                levels = np.linspace(0, min(extend, 1), int(10 * min(extend, 1))) ** 2
                if extend > 1:
                    levels2 = np.linspace(1, extend, int(5 * (extend - 1)))[1:] ** 2
                    levels = np.hstack([levels, levels2])

                ax.contour(*transform((X, Y)), H, levels=levels, **kwargs)

                ax.grid(False)

            # autoscale to separatrix
            if autoscale:
                t = 1 if autoscale is True else autoscale
                X = np.array((-2, -2, 1 - 3 * t, 1 + 3 * t, 1 - 3 * t, 1 + 3 * t)) / 2
                Y = np.array((-2 * t, 2 * t, 1 + t, 1 - t, -1 - t, -1 + t)) * 3**0.5 / 2
                x, y = transform((h * X, h * Y))
                self._autoscale(ax, data=np.transpose((x, y)), reset=True)


__all__ = PUBLIC_SECTION_END()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Color definitions

Colors from https://arxiv.org/abs/2107.02270

"""

__author__ = "Philipp Niedermayer"
__contact__ = "eltos@outlook.de"
__date__ = "2022-09-06"


import matplotlib as mpl
from .util import *


PUBLIC_SECTION_BEGIN()


petroff_colors = [
    "#3f90da",
    "#ffa90e",
    "#bd1f01",
    "#94a4a2",
    "#832db6",
    "#a96b59",
    "#e76300",
    "#b9ac70",
    "#717581",
    "#92dadd",
]
"""List of 10 accessible and aesthetic colors selected by M. A. Petroff"""

cmap_petroff = mpl.colors.ListedColormap(petroff_colors, "petroff")
"""Colormap with 10 distinct colors"""
cmap_petroff.set_under("none")
cmap_petroff.set_over("none")

cmap_petroff_gradient = mpl.colors.LinearSegmentedColormap.from_list(
    "petroff_gradient", [petroff_colors[i] for i in (9, 0, 4, 2, 6, 1)]
)
"""Gradient colormap"""
cmap_petroff_gradient.set_under(petroff_colors[3])
cmap_petroff_gradient.set_over(petroff_colors[7])

cmap_petroff_bipolar = mpl.colors.LinearSegmentedColormap.from_list(
    "petroff_bipolar", [petroff_colors[i] for i in (2, 6, 1, 3, 9, 0, 4)]
)
"""Bipolar colormap"""
cmap_petroff_bipolar.set_under(petroff_colors[5])
cmap_petroff_bipolar.set_over(petroff_colors[8])

cmap_petroff_cyclic = mpl.colors.LinearSegmentedColormap.from_list(
    "petroff_cyclic", [petroff_colors[i] for i in (3, 9, 0, 4, 2, 6, 1, 3)]
)
"""Cyclic colormap"""


def make_unicoloured_cmap(color):
    """Make a linear colormap of a given color from transparent to black

    Args:
        color (str | tuple): Color as hex string or RGB tuple

    Returns:
        Colormap
    """
    return mpl.colors.LinearSegmentedColormap.from_list(
        "gradient",
        (
            (0.000, (1.000, 1.000, 1.000, 0)),
            (0.400, color),
            (1.000, (0.000, 0.000, 0.000, 1)),
        ),
    )


__all__ = PUBLIC_SECTION_END()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Utility methods for accelerator physics"""

__author__ = "Philipp Niedermayer"
__contact__ = "eltos@outlook.de"
__date__ = "2022-11-15"


import inspect
import re

import numpy as np
import scipy.signal
import matplotlib as mpl
import matplotlib.collections
import matplotlib.patches
import matplotlib.cbook
import matplotlib.text
import matplotlib.lines

try:
    import pandas as pd
except ImportError:
    # pandas is an optional dependency
    pd = None


def PUBLIC_SECTION_BEGIN():
    """Starting from here, collect objects to be included in file's `__all__` magic

    :nodoc:
    """
    g = inspect.stack()[1].frame.f_globals
    g["__private_names"] = set(g.keys()).union(["__private_names"])


def PUBLIC_SECTION_END():
    """Finish object collection and return global names for the `__all__` magic

    :nodoc:
    """
    g = inspect.stack()[1].frame.f_globals
    public = list(set(g.keys()).difference(g["__private_names"]))
    return public[:]  # copy needed for it to work with sphinx autoapi


VOID = object()
AUTO = object()


c0 = 299792458
"""Speed of light in m/s"""


def val(obj):
    """Return the value if this is an array of size 1, object otherwise"""
    if np.size(obj) == 1:
        return np.array(obj).item()
    return obj


#
def ieee_mod(values, m):
    """Return the IEEE remainder (in range -x/2 .. x/2)"""
    return np.mod(values + m / 2, m) - m / 2


def get(obj, value, default=VOID):
    """Get value from object

    Tries to get the value using attributes and indices,
    and handles special objects like pandas data frames.

    Args:
        obj (Any): Object to get data from. Can also be a tuple of objects.
        value (str): Name of attribute, index, column etc. to get
        default (Any): Default value to return. By default an exception is raised

    Returns:
        Any: Value of the object

    Raises:
        AttributeError: if object does not provide the value and no default was specified
    """
    if pd and isinstance(obj, pd.DataFrame):
        return val(obj[value].values)
    if type(obj) == tuple:
        for o in obj:
            if (v := get(o, value, None)) is not None:
                return v
    try:
        return val(getattr(obj, value))
    except:
        try:
            return val(obj[value])
        except:
            if default is not VOID:
                return default
    raise AttributeError(f"{obj} does not provide an attribute or index '{value}'")


def defaults(kwargs, /, **default_kwargs):
    """Return keyword arguments with defaults

    Returns a union of keyword arguments, where `kwargs` take precedence over `default_kwargs`.

    Args:
        kwargs (dict | None): keyword arguments (overwrite defaults)
        default_kwargs: default keyword arguments
    """
    return dict(default_kwargs, **(kwargs or {}))


def defaults_for(alias_provider, kwargs, /, **default_kwargs):
    """Return normalized keyword arguments with defaults

    Returns a union of keyword arguments, where `kwargs` take precedence over `default_kwargs`.
    All keyword arguments are normalized beforehand via :meth:`matplotlib.cbook.normalize_kwargs`.

    Args:
        alias_provider (str | dict | class | artist): alias provider for :meth:`matplotlib.cbook.normalize_kwargs`
        kwargs (dict | None): keyword arguments (overwrite defaults)
        default_kwargs: default keyword arguments
    """
    # resolve aliases
    if isinstance(alias_provider, str):
        alias_provider = dict(
            text=mpl.text.Text,
            plot=mpl.lines.Line2D,
            fill_between=mpl.patches.Polygon,
            scatter=mpl.collections.Collection,
            hexbin=mpl.collections.PolyCollection,
            vlines=mpl.collections.LineCollection,
        )[alias_provider]
    if alias_provider == mpl.patches.Polygon and kwargs is not None and "c" in kwargs:
        kwargs["color"] = kwargs.pop("c")
    kwargs = mpl.cbook.normalize_kwargs(kwargs, alias_provider)
    default_kwargs = mpl.cbook.normalize_kwargs(default_kwargs, alias_provider)

    # handle custom values
    def apply_custom_to(kwargs):
        if (ls := kwargs.get("linestyle")) and isinstance(ls, str):
            if match := re.match(r"-(\d+)", ls):
                n = int(match.group(1))
                kwargs["linestyle"] = (0, [] if n == 0 else [1, 1.6] * n + [48.4 - 2.6 * n, 1.6])
            if match := re.match(r"-\.(\d+)", ls):
                n = int(match.group(1))
                kwargs["linestyle"] = (0, [1, 1.6] * (n + 1) + [6.4, 1.6])

    apply_custom_to(kwargs)
    apply_custom_to(default_kwargs)
    # apply with defaults
    return defaults(kwargs, **default_kwargs)


def flattened(lists):
    """Flatten a list of nested lists recursively"""
    if hasattr(lists, "__iter__") and not isinstance(lists, str):
        return [item for sublist in lists for item in flattened(sublist)]
    return [lists]


class AttrDict(dict):
    """Dict which allows accessing values via key attributes"""

    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


#    def update_recursive(self, other):
#        for k, v in other.items():
#            if k in self and isinstance(self[k], collections.abc.Mapping):
#                self[k] = AttrDict(self[k])
#                self[k].update_recursive(v)
#            else:
#                self[k] = v


def average(*data, n=100, function=np.mean, logspace=False, keepdim=False):
    """Average the data

    Applies the function to n subsequent points of the data (along last axis) to yield one point in the output

    Args:
        data (np.ndarray): the data to average over
        n (int | floor): number of subsequent datapoints of intput to average into one point in the output.
            If the input size is not a multiple of n, the data will be clipped.
            In case of logspace scaling, this can be a flost (usually 1 < n < 2 gives good results)
        function (function): averaging function to apply to last axis of input data. Defaults to np.mean
        logspace (bool): If true, average N subsequent points where N is adjusted along the data to yield equal window sizes on a log scale
        keepdim (bool): If true, repeat averaged data to keep the diemension of the input array

    Returns:
        averaged data
    """
    result = []
    for d in data:
        if not logspace:
            # linear space: work on every subsequent n samples
            w = int(d.shape[-1] / n)
            new = function(d[..., : w * n].reshape(*d.shape[:-1], w, n), axis=-1)
            if keepdim:
                new = np.repeat(new, n, axis=-1)
                new = np.pad(
                    new,
                    [(0, 0)] * (new.ndim - 1) + [(0, d.shape[-1] - new.shape[-1])],
                    constant_values=np.nan,
                )
        else:
            # log space: work on every subsequent n**i samples
            N = n ** np.arange(1, np.log(d.shape[-1]) / np.log(n))
            N = np.unique(np.hstack(([0], N[N < d.shape[-1]].astype("int"))))
            new = np.nan * (
                np.empty_like(d) if keepdim else np.empty([*d.shape[:-1], N.size - 1])
            )
            for i, (n0, n1) in enumerate(zip(N[:-1], N[1:])):
                new[..., slice(n0, n1) if keepdim else i] = v = function(d[..., n0:n1])

        result.append(new)

    return result[0] if len(result) == 1 else result


def smooth(*data, n):
    """Smooth the data over n consecutive bins

    Args:
        *data (np.array): The data
        n (int): Number of items to smooth over
    Returns
         (np.array | list[np.array]): The smooth array(s) with same shape as the original data
             (but first and last n/2 values will be np.nan).

    """
    if n:
        # preserves shape but first and last n/2 values will be np.nan
        data = [scipy.signal.savgol_filter(d, n, 0, mode="constant", cval=np.nan) for d in data]
    return data[0] if len(data) == 1 else data


def evaluate_expression_wrapper(expression, key, data):
    """Evaluate the expression wrapper"""

    methods = dict(np=np, smooth=smooth, average=average, offset=lambda x, o: x + o)

    try:
        return eval(expression, methods, {key: data})
    except Exception as e:
        import inspect, sys

        print(f"Error evaluating exprssion `{expression}`", file=sys.stderr)
        print(f"Reason: {e}", file=sys.stderr)
        print(f"Supported methods:", file=sys.stderr)
        for k, v in methods.items():
            if k in ("self", "data") or k.startswith("_"):
                continue
            print(f"  {k}{inspect.signature(v)}", file=sys.stderr)
        raise


def binned_data(
    values, *, what=None, n=None, dv=None, v_range=None, moments=1, make_n_power_of_two=False
):
    """Get histogrammed data with equally spaced bins

    From the non-equally distributed values, a histogram or timeseries with equally
    spaced bins is derived. The parameter ``what`` determines what is returned for the timeseries.
    By default (what=None), the number of particles for each bin is returned.
    Alternatively, a particle property can be passed as array, in which case that property is averaged
    over all particles of the respective bin (or 0 if no particles are within a bin).
    It is also possible to specify the moments to return, i.e. the power to which the property is raised
    before averaging. This allows to determine mean (1st moment, default) and variance (difference between
    2nd and 1st moment) etc. To disable averaging, pass None as the moment

    Args:
        values (np.ndarray): Array of particle data, e.g. timestamps.
        n (int | None): Number of bins. Must not be used together with `dv`.
        dv (float | None): Bin width. Must not be used together with n.
        v_range (tuple[float] | None): Tuple of (min, max) values to consider.
            If None, the range is determined from the data.
        what (np.ndarray | None): Array of associated data or None. Must have same shape as values. See above.
        moments (int | list[int | None] | None): The moment(s) to return for associated data if what is not None. See above.
        make_n_power_of_two (bool): If true, ensure that the number of bins is a power of two by rounding up.
            Useful to increase performance of calculating FFTs on the timeseries data.

    Returns:
        tuple: The histogram or timeseries as tuple
            ``(v_min, dv, *counts_or_what)`` where
            ``v_min`` is the start value of the histogram or timeseries data,
            ``dv`` is the bin width and
            ``*counts_or_what`` are the values of the histogram or timeseries as an array of length n
            for each moment requested.
    """

    v_min = np.min(values) if v_range is None or v_range[0] is None else v_range[0]
    v_max = np.max(values) if v_range is None or v_range[1] is None else v_range[1]

    if n is not None and dv is None:
        # number of bins requested, adjust bin width accordingly
        if make_n_power_of_two:
            n = 1 << (n - 1).bit_length()
        dv = (v_max - v_min) / n
    elif n is None and dv is not None:
        # bin width requested, adjust number of bins accordingly
        n = int(np.ceil((v_max - v_min) / dv))
        if make_n_power_of_two:
            n = 1 << (n - 1).bit_length()
            dv = (v_max - v_min) / n  # adjust bin width to match new n
    else:
        raise ValueError(f"Exactly one of n or dt must be specified, but got n={n} and dt={dv}")

    # Note: The code below was optimized to run much faster than an ordinary np.histogram,
    # which quickly slows down for large number of bins as required for FFT calculation.
    # Benchmark: Binning of a data array of size=1000000 into 100000000 bins
    #     np.histogram(data, bins=100000000) takes 6.21 s ± 374 ms
    #     binned_data(data, n=100000000) takes 93.9 ms ± 2.8 ms
    # If you intend to change something here, make sure to benchmark it!

    # count timestamps in bins
    bins = np.floor((values - v_min) / dv).astype(int)
    # bins are i*dt <= t < (i+1)*dt where i = 0 .. n-1
    mask = (bins >= 0) & (bins < n)  # ignore times outside range
    bins = bins[mask]
    # count particles per time bin
    counts = np.bincount(bins, minlength=n)[:n]

    if what is None:
        # Return particle counts
        return v_min, dv, counts

    else:
        # Return 'what' averaged
        result = [v_min, dv]
        if isinstance(moments, int) or moments is None:
            moments = [moments]
        for m in moments:
            v = np.zeros(n)
            # sum up 'what' for all the particles in each bin
            power = m if m is not None else 1
            np.add.at(v, bins, what[mask] ** power)
            if m is not None:
                # divide by particle count to get mean (default to 0)
                v[counts > 0] /= counts[counts > 0]
            result.append(v)
        return result


def element_strength(element, n):
    """Get knl strength of element"""
    knl = 0
    if knl == 0 and hasattr(element, f"k{n}") and hasattr(element, "length"):
        knl = getattr(element, f"k{n}") * element.length
    if knl == 0 and hasattr(element, "knl") and n <= element.order:
        knl = element.knl[n]
    return knl


def normalized_coordinates(x, px, twiss, xy, delta=0):
    """Convert physical to normalized coordinates

    Args:
        x (float): Physical x-coordinate in m
        px (float): Physical px-coordinate in rad
        twiss (Any): Object or dict with twiss information in m and rad
        xy (str): Plane. Either "x" or "y".
        delta (float): Momentum deviation to account for dispersive orbit.

    Returns:
        tuple: Tuple of normalized coordinates
            (X, Px) in (m^(1/2), m^(1/2))
    """
    if twiss is None:
        raise ValueError("Cannot calculate normalized coordinates when twiss parameter is None")
    # substract (dispersive) closed orbit
    x = x - get(twiss, xy) - delta * get(twiss, "d" + xy)
    px = px - get(twiss, "p" + xy) - delta * get(twiss, "dp" + xy)
    # apply Floquet transform
    alf, bet = get(twiss, "alf" + xy), get(twiss, "bet" + xy)
    X = x / bet**0.5
    Px = alf * x / bet**0.5 + px * bet**0.5
    return X, Px


def denormalized_coordinates(X, Px, twiss, xy, delta=0):
    """Convert normalized to physical coordinates

    Args:
        X (float): Normalized X-coordinate in m^(1/2)
        Px (float): Normalized Px-coordinate in m^(1/2)
        twiss (Any): Object or dict with local twiss information in m and rad
        xy (str): Plane. Either "x" or "y".
        delta (float): Momentum deviation to account for dispersive orbit.

    Returns:
        tuple: Tuple of physical coordinates
            (x, px) in (m, rad)
    """
    # apply Floquet transform
    alf, bet = get(twiss, "alf" + xy), get(twiss, "bet" + xy)
    x = X * bet**0.5
    px = -alf * X / bet**0.5 + Px / bet**0.5
    # add (dispersive) closed orbit
    x = x + get(twiss, xy) + delta * get(twiss, "d" + xy)
    px = px + get(twiss, "p" + xy) + delta * get(twiss, "dp" + xy)
    return x, px


def virtual_sextupole(line, particle_ref=None, *, verbose=False):
    """Determine virtual sextupole strength from the line

    The normalized strenght is defined as S = -1/2 * betx^(3/2) * k2l

    The implementation considers only normal sextupole components.

    Args:
        line (xtrack.Line): Line with element_dict and twiss method
        particle_ref (xpart.Particles): Reference particle. Defaults to reference particle of tracker.
        verbose (bool): If True, print information on sextupoles

    Returns:
        tuple: Tuple (S, mu)
            with normalized strength in m^(-1/2) and phase in rad/2pi
    """

    # for backwards compatibility for old xsuite versions
    line = line.line if hasattr(line, "line") else line

    # find sextupoles
    sextupoles, k2l = [], []
    for name, el in line.element_dict.items():
        if element_k2l := element_strength(el, 2):
            sextupoles.append(name)
            k2l.append(element_k2l)

    # twiss at sextupoles
    tw = line.twiss(method="4d", particle_ref=particle_ref, at_elements=sextupoles)
    betx, mux = tw.betx, tw.mux

    # determine virtual sextupole
    Sn = -1 / 2 * betx ** (3 / 2) * k2l
    vectors = Sn * np.exp(3j * mux * 2 * np.pi)
    V = np.sum(vectors)
    S = np.abs(V)
    mu = np.angle(V) / 3 / 2 / np.pi

    if verbose:
        print("Sextupoles")
        data = dict(
            name=sextupoles,
            k2l=k2l,
            betx=betx,
            mux=mux,
            dx=tw.dx,
            S_abs=np.abs(vectors),
            S_deg=np.rad2deg(np.angle(vectors)),
        )
        head = data.keys()
        info = [[v if type(v) is str else f"{v:.5f}" for v in data[h]] for h in head]
        colw = [max(len(h), *[len(v) for v in val]) for h, val in zip(head, info)]
        info = ["  ".join([v.rjust(c) for v, c in zip(val, colw)]) for val in (head, *zip(*info))]
        print("  " + "\n  ".join(info))
        print("  " + "-" * len(info[0]))
        print("  Virtual sextupole:", f"S = {S:g} m^(-1/2) at mu = {mu:g} rad/2pi")

    return S, mu


def hamiltonian_kobayashi(X, Px, S, mu, twiss, xy="x", delta=0, *, normalized=False):
    """Calculate the kobayashi hamiltonian

    Args:
        X (float): Normalized X-coordinate in m^(1/2)
        Px (float): Normalized Px-coordinate in m^(1/2)
        S (float): Normalized sextupole strength in m^(-1/2)
        mu (float): Sextupole phase in rad/2pi
        twiss (Any): Object or dict with local twiss information in m and rad
        xy (str): Plane. Either "x" or "y".
        delta (float): Momentum deviation to account for dispersive orbit.
        normalized (bool): If true, return value of hamiltonian divided by value at separatrix.

    Returns:
        float: Value of the hamiltonian
            (possibly normalized)

    """

    # Tune distance
    q = get(twiss, "q" + xy)
    q = q + delta * get(twiss, "dq" + xy)  # chromatic tune shift
    r = np.round(3 * q) / 3  # next 3rd order resonance
    d = q - r  # distance to resonance
    # h = 4 * np.pi * d / S  # size of stable striangle (inscribed circle)

    # rotate coordinates to account for phase advance
    dmu = 2 * np.pi * (mu - get(twiss, "mu" + xy))  # phase advance to virtual sextupole
    rotation = np.array([[+np.cos(dmu), np.sin(dmu)], [-np.sin(dmu), np.cos(dmu)]])
    (X, Px) = np.tensordot(rotation, (X, Px), 1)

    # calculate hamiltonian
    H = 3 * np.pi * d * (X**2 + Px**2) + S / 4 * (3 * X * Px**2 - X**3)
    if normalized:
        Hsep = (4 * np.pi * d) ** 3 / S**2
        H = H / Hsep

    return H


def iter_elements(line, *, s=None, mask_nearest=False):
    """Iterate over elements in line

    Args:
        line (xtrack.Line): Line of elements.
        s (np.array | None): Optional array of s positions.
          If not None, returns an additional s-position-mask for each element.
        mask_nearest (bool): If true, zero-length elements will be masked at the nearest s-coordinate of the s array
          to avoid them being skipped if their location does not coincide exactly with any value in the s-array.
          If s is None, this has no effect.

    Yields:
        (name, element, s_from, s_to, [s_mask]): Name, object, start position,
          end position and optional position mask for each element. The mask is
          omitted if s=None (see arguments).

    """
    el_s0 = line.get_s_elements("upstream")
    el_s1 = line.get_s_elements("downstream")
    smax = line.get_length()
    for name, el, s0, s1 in zip(line.element_names, line.elements, el_s0, el_s1):
        if s0 == s1:  # thin lense located at element center
            if hasattr(el, "length"):
                s0, s1 = (s0 + s1 - el.length) / 2, (s0 + s1 + el.length) / 2
        if s is None:
            yield name, el, s0, s1
        else:
            if s0 == s1:
                if mask_nearest:
                    # zero-length element -> round to nearest s
                    mask = s == s[np.argmin(np.abs(s - s0))]
                else:
                    mask = s == s0 % smax
            elif 0 <= s0 <= smax:
                mask = (s >= s0) & (s < s1)
            else:  # handle wrap around
                mask = (s >= s0 % smax) | (s < s1 % smax)
            yield name, el, s0, s1, mask


def apertures(line, at_s=None):
    """Determine aperture

    Args:
        line (xtrack.Line): Line with element_dict
        at_s (np.ndarray | None): return apertures for these s locations in m

    """

    if at_s is None:
        s = np.unique([(s0, s1) for _, _, s0, s1 in iter_elements(line)])
    else:
        s = np.array(at_s)

    result = dict(
        s=s,
        min_x=np.full_like(s, np.nan),
        max_x=np.full_like(s, np.nan),
        min_y=np.full_like(s, np.nan),
        max_y=np.full_like(s, np.nan),
    )
    for name, el, s0, s1, mask in iter_elements(line, s=s):
        for xy in "xy":
            # find limits
            mi = getattr(el, f"min_{xy}", None)
            ma = getattr(el, f"max_{xy}", None)
            if mi is None and ma is not None:
                mi = -ma
            # handle shift
            if mi is not None:
                mi += getattr(el, f"shift_{xy}", 0)
                result[f"min_{xy}"][
                    mask & ~(result[f"min_{xy}"] > mi)
                ] = mi  # comparison inverted to handle NaN
            if ma is not None:
                ma += getattr(el, f"shift_{xy}", 0)
                result[f"max_{xy}"][mask & ~(result[f"max_{xy}"] < ma)] = ma

    if at_s is None:
        # only return s where there is an aperture
        mask = np.any(
            [np.isfinite(result[f"{m}_{xy}"]) for m in ("min", "max") for xy in "xy"], axis=0
        )
        for key in result:
            result[key] = result[key][mask]

    return AttrDict(result)



import os
import shutil
import base64
import re
import warnings
from unittest import TestCase
from copy import deepcopy
from tempfile import NamedTemporaryFile

import pytest
from testbook import testbook
from matplotlib.testing.compare import compare_images

dir = "examples"
if not os.path.exists(dir):
    dir = os.path.join("..", dir)


class NotebookTester:
    def __init__(self, nb):
        self.nb = nb

    def execute(self, *, check_outputs=False):
        with testbook(self.nb) as tb:
            for i, cell in enumerate(tb.cells):
                tags = cell.get("metadata", {}).get("tags", [])
                ref_cell = deepcopy(cell)

                # execute
                cell = tb.execute_cell(i)

                if check_outputs:
                    try:
                        self.compare_cell_outputs(ref_cell, cell)
                    except:
                        print(
                            f"\nError in {self.nb} cell [{ref_cell.get('execution_count')}] (ID:{ref_cell.get('id')}): Output not reproduced."
                        )
                        print(
                            "Cell source:\n    "
                            + "\n    ".join(ref_cell.get("source", "").split("\n"))
                        )
                        raise

                # yield i, tags, ref_cell, cell, tb

    @classmethod
    def compare_cell_outputs(cls, reference_cell, actual_cell):
        """Compare cell outputs unless remove-tags exist"""

        # ignore cells with remove-output tag
        tags = reference_cell.get("metadata", {}).get("tags", [])
        if "remove-output" in tags or "skip-test" in tags:
            return True  # skip completely

        # ignore stderr/stdout cell outputs if cell tagged with remove-stderr or remove-stdout
        def keep(out):
            if out.get("output_type") == "stream" and f"remove-{out.get('name')}" in tags:
                warnings.warn(
                    f"Not checking {out.get('name')} output of cell [{reference_cell.get('execution_count')}]"
                    f" tagged with remove-{out.get('name')}."
                )
                return False
            return True

        # Make sure these are lists, since filter object can be iterated over only once!
        reference_outputs = list(filter(keep, reference_cell.get("outputs", [])))
        actual_outputs = list(filter(keep, actual_cell.get("outputs", [])))

        try:
            # compare the filtered outputs
            for i, ref_out in enumerate(reference_outputs):
                output_type = ref_out.get("output_type")

                try:
                    act_out = actual_outputs[i]
                except IndexError:
                    raise ValueError(
                        f"Expected {len(reference_outputs)} cell outputs, but only {len(actual_outputs)} found"
                    )

                if output_type == "stream":  # compare plain text
                    cls.compare_outputs_text(ref_out.get("text"), act_out.get("text"))

                else:
                    output_data = ref_out.get("data", {})

                    for mime, rule in {
                        "image/png": cls.compare_outputs_image,
                        "text/markdown": cls.compare_outputs_markup,
                        "text/html": cls.compare_outputs_html,
                    }.items():
                        if mime in output_data:
                            rule(output_data.get(mime), act_out.get("data", {}).get(mime))
                            break
                    else:  # unknown output type
                        raise NotImplementedError(
                            f"No rule to compare output type '{ref_out.get('output_type')}' with data '{ref_out.get('data', {}).keys()}'. \n"
                            "Currently only 'image/png', 'text/markdown' and 'text/html' are supported."
                        )

        except:
            for label, outputs in (
                ("Reference outputs:", reference_outputs),
                ("Actual outputs:", actual_outputs),
            ):
                print(label)
                for o in outputs:
                    print(o.get("output_type"))
                    print(o.get("text", ""))
                    print()

            raise

    @classmethod
    def compare_outputs_text(cls, reference_text, actual_text):
        """Plain text comparison"""
        test = TestCase()
        test.assertEqual(reference_text, actual_text)

    @classmethod
    def compare_outputs_markup(cls, ref_str, act_str):
        """Markup text comparison (ignoring whitespace changes)"""
        ref_str, act_str = [re.sub(r"\s+", " ", _) for _ in (ref_str, act_str)]
        TestCase().assertEqual(ref_str, act_str)

    @classmethod
    def compare_outputs_html(cls, ref_str, act_str):
        """HTML text comparison"""
        TestCase().assertEqual(ref_str, act_str)

    @classmethod
    def compare_outputs_image(cls, ref_img, act_img, *, tol=2):
        """Compare two base64 encoded images"""
        assert act_img, f"Expected image/png cell output, but {act_img} found"

        # save to temp file and compare
        with NamedTemporaryFile(suffix=".png") as rf, NamedTemporaryFile(suffix=".png") as af:
            rf.write(base64.b64decode(ref_img)) and rf.flush()
            af.write(base64.b64decode(act_img)) and af.flush()

            if diff := compare_images(rf.name, af.name, tol, True):
                os.makedirs("tests/output/", exist_ok=True)
                diff_image = shutil.copy(diff["diff"], "tests/output/")
                raise ValueError(
                    f"Image output differs from excpected output: RMS value {diff['rms']:g} > {diff['tol']:g} exceeds tolerance."
                    f"Diff image written to {diff_image}."
                )


def test_animations():
    t = NotebookTester(f"{dir}/animations.ipynb")
    t.execute(check_outputs=False)  # no good way of checking animation result


def test_colors():
    t = NotebookTester(f"{dir}/colors.ipynb")
    t.execute(check_outputs=True)


def test_hamiltonians():
    t = NotebookTester(f"{dir}/hamiltonians.ipynb")
    t.execute(check_outputs=True)


def test_line():
    t = NotebookTester(f"{dir}/line.ipynb")
    t.execute(check_outputs=True)


def test_phasespace():
    t = NotebookTester(f"{dir}/phasespace.ipynb")
    t.execute(check_outputs=True)


def test_concepts():
    t = NotebookTester(f"{dir}/concepts.ipynb")
    t.execute(check_outputs=True)


def test_timestructure():
    t = NotebookTester(f"{dir}/timestructure.ipynb")
    t.execute(check_outputs=True)


def test_twiss():
    t = NotebookTester(f"{dir}/twiss.ipynb")
    t.execute(check_outputs=True)


def test_utilities():
    t = NotebookTester(f"{dir}/utilities.ipynb")
    t.execute(check_outputs=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Philipp Niedermayer"
__contact__ = "eltos@outlook.de"


__version__ = "0.11.12"


# expose the following in global namespace
from .colors import *
from .line import *
from .particles import *
from .phasespace import *
from .properties import *
from .timestructure import *
from .twiss import *
from .util import AUTO

# allow usage of xplt.mpl.* without importing matplotlib
import matplotlib as mpl


# Deprecated, for backwards compatibility
class TimeVariationPlot(SpillQualityPlot):
    """
    .. deprecated:: 0.8
        Use :class:`xplt.SpillQualityPlot` instead.
    """

    pass


class TimeVariationScalePlot(SpillQualityTimescalePlot):
    """
    .. deprecated:: 0.8
        Use :class:`xplt.SpillQualityTimescalePlot` instead.
    """

    pass


from . import hooks as _hooks

_hooks.try_register_hooks()

import matplotlib.style as _mpl_style


def apply_style():
    """Apply xplt's matplotlib style sheet and update rcParams"""
    try:
        _mpl_style.use("xplt.xplt")
    except IOError:
        import matplotlib, shutil, os

        # for matplotlib versions prior to 3.7: copy the style file to the config directory
        dir = os.path.join(matplotlib.get_configdir(), "stylelib")
        os.makedirs(dir, exist_ok=True)
        print("Installing xplt style in", dir)
        shutil.copy(os.path.join(os.path.dirname(__file__), "xplt.mplstyle"), dir)
        matplotlib.style.reload_library()
        matplotlib.style.use("xplt")

    # import pint.formatting as _pint_formatting
    # _pint_formatting.format_default = "X"  # use explicit format instead!



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Matplotlib setup"""

__author__ = "Philipp Niedermayer"
__contact__ = "eltos@outlook.de"
__date__ = "2022-09-06"

import types
import warnings
from .colors import (
    cmap_petroff,
    cmap_petroff_gradient,
    cmap_petroff_bipolar,
    cmap_petroff_cyclic,
    petroff_colors,
)


def try_register_hooks():
    try:
        register_matplotlib_options()
    except Exception as e:
        warnings.warn(f"Failed to register color names with matplotlib: {e}", RuntimeWarning)
        pass

    try:
        register_pint_options()
    except Exception as e:
        warnings.warn(f"Failed to register formatters with pint: {e}", RuntimeWarning)
        pass


def register_matplotlib_options():
    """Register default options for matplotlib"""
    import matplotlib as mpl

    # register named colors
    for i, color in enumerate(petroff_colors):
        mpl.colors.get_named_colors_mapping().update({f"p{i}": color})
        mpl.colors.get_named_colors_mapping().update({f"pet{i}": color})

    # register named colormaps
    for cmap in (
        cmap_petroff,
        cmap_petroff_gradient,
        cmap_petroff_bipolar,
        cmap_petroff_cyclic,
    ):
        mpl.colormaps.register(cmap=cmap)
        cmap_r = cmap.reversed()
        cmap_r.name = cmap.name + "_r"
        mpl.colormaps.register(cmap=cmap_r)

    # register custom scales
    from .base import DiscontinuousLinearScale

    mpl.scale.register_scale(DiscontinuousLinearScale)


def register_pint_options():
    """Register default options for pint"""
    import pint

    @pint.register_unit_format("X")
    def format_latex(unit, registry, **options):
        """Slightly modified latex formatter"""
        preprocessed = {
            r"\mathrm{{{}}}".format(u.replace("_", r"\_")): p for u, p in unit.items()
        }
        formatted = pint.formatter(
            preprocessed.items(),
            {},
            as_ratio=False,  # changed
            single_denominator=True,
            product_fmt=r" \cdot ",
            division_fmt=r"\frac[{}][{}]",
            power_fmt="{}^[{}]",
            parentheses_fmt=r"\left({}\right)",
            **options,
        )
        return formatted.replace("[", "{").replace("]", "}").replace("^{0.5}", "^{1/2}")


## Restrict star imports to local namespace
__all__ = [
    name
    for name, thing in globals().items()
    if not (name.startswith("_") or isinstance(thing, types.ModuleType))
]
